import tkinter as tk
from tkinter import font

from psopackage.graphics.colors import Color
from psopackage.graphics.fonts import FontName
from psopackage.graphics.mainMenu.bottomButton import BottomButton
from psopackage.graphics.mainMenu.popUpFrame import PopUpFrame

class BottomFrame(tk.Frame):
    """ 
    Manages the bottom frame of the GUI that contains the buttons for the info and help pop up frames.
    """
    def __init__(self, master_frame, menu_frame: tk.Frame, window_width: int, window_height, title_height: int, bottom_frame_height: int, program_version: str):
        super().__init__(menu_frame, bg=Color.test3_bg)
        self.__window_width: int = window_width
        self.__window_height: int = window_height
        self.__height: int = bottom_frame_height
        self._pop_up_frame_height: int = self.__window_height - title_height - self.__height

        info_image: tk.PhotoImage = tk.PhotoImage(file="graphics/assets/info.png").subsample(3)
        info_active_image: tk.PhotoImage = tk.PhotoImage(file="graphics/assets/info-active.png").subsample(3)
        self.__info_button: BottomButton = BottomButton(master_frame, self, image=info_image, active_image=info_active_image, name="info", pop_up_frame_height=self._pop_up_frame_height, pop_up_frame_width=self.__window_width, pop_up_frame_y=title_height, pop_up_text="This is a program to simulate the PSO algorithm. It's an optimization algorithm, that its purpose is to find the minimum of a function. As of Version 1.0.0, you manually select the function to be optimized, having four options, Sphere, Booth, Rastrigin and Goldstein-Price.")
        help_image: tk.PhotoImage = tk.PhotoImage(file="graphics/assets/help.png").subsample(3)
        help_active_image: tk.PhotoImage = tk.PhotoImage(file="graphics/assets/help-active.png").subsample(3)
        self.__help_button: BottomButton = BottomButton(master_frame, self, image=help_image, active_image=help_active_image, name="help", pop_up_frame_height=self._pop_up_frame_height, pop_up_frame_width=self.__window_width, pop_up_frame_y=title_height, pop_up_text="First, you click on create optimization, manually type the number of iterations, particles or the coefficients needed for the algorithm, and then you click Run Optimization. You then can go back to the menu and click Select optimization to view the results of it, displaying the particles graphed on each iteration. ALternatively, you can directly click View Results and it will redirect you to the select menu.")
        self._pop_up_frames: list[PopUpFrame] = [self.__help_button.pop_up_frame, self.__info_button.pop_up_frame] # ! Could be immutable actually
        
        self.__version_label: tk.Label = tk.Label(
            self,
            text=f"V {program_version}",
            bg=Color.bottom_button_bg,
            fg=Color.bottom_button_fg,
            font=font.Font(family=FontName.label, size=10, weight="bold")
            )

    def display(self):     
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.columnconfigure(2, weight=1)
        self.__info_button.display(self._pop_up_frames, row=0, column=0, sticky="nsew")
        self.__help_button.display(self._pop_up_frames, row=0, column=1, sticky="nsew")
        self.__version_label.grid(row=0, column=2, sticky="nsew")
        self.place(x=0, y=self.__window_height - self.__height, width=self.__window_width, height=self.__height)



# ! An image class could also be made! There could really be all tk's widgets' classes!
# * This could make a future implementation with ttk or any other module much easier
import numpy as np
import tkinter as tk

from psopackage.graphics.colors import Color
from psopackage.optimization import Optimization
from psopackage.graphics.selectMenu.viewButton import ViewButton

class OptimizationFrame:
    def __init__(self, master: tk.Frame, optimization: Optimization, forget_select_menu: callable, initialize_window: callable, change_menu: callable, width: int, height: int, separation: int, scrollbar_width: int, frame_index: int) -> None:
        self.root: tk.Frame = tk.Frame(master, width=width, height=height)
        self.__width: int = width
        self.__height: int = height
        self.__scrollbar_width: int = scrollbar_width
        self.__separation: int = separation
        self.__index: int = frame_index
        self.__widget_parameters: dict = {
            "master": self.root,
            "bg": Color.select_label_optim_bg,
            "fg": Color.select_label_optim_fg
        }
        # TODO: Try to add the self.root in the dictionary
        # * Seems to be working just fine! Could be implemented in other menus if the master declaration is repeated many times.
        self.__name_label: tk.Label = tk.Label(text=f"Optimization {optimization.index}", **self.__widget_parameters)
        self.__function_label: tk.Label = tk.Label(text=optimization.function_choice, **self.__widget_parameters)
        self.__dimensions_label: tk.Label = tk.Label(text=f"Dimensions: {optimization.dimensions}", **self.__widget_parameters)
        self.__minima_indicator_label: tk.Label = tk.Label(text=f"Minima: ", **self.__widget_parameters)
        minima_coordinates: np.ndarray = np.round(optimization.swarm.get_gbest().get_coordinates(), 3)
        self.__minima_value_label: tk.Label = tk.Label(text=minima_coordinates, **self.__widget_parameters)
        self.__cognitive_coefficient_label: tk.Label = tk.Label(text=f"c1: {optimization.swarm.get_cognitive_coefficient()}", **self.__widget_parameters)
        self.__num_particles_label: tk.Label = tk.Label(text=f"N. of particles: {optimization.swarm.get_particle_amount()}", **self.__widget_parameters)
        self.__social_coefficient_label: tk.Label = tk.Label(text=f"c2: {optimization.swarm.get_social_coefficient()}", **self.__widget_parameters)

        view_button_image: tk.PhotoImage = tk.PhotoImage(file="graphics/assets/preview.png").subsample(4)
        view_button_active_image: tk.PhotoImage = tk.PhotoImage(file="graphics/assets/preview-active.png").subsample(4)
        self.view_button: ViewButton = ViewButton(master=self.root, image=view_button_image, active_image=view_button_active_image, forget_select_menu=forget_select_menu, initialize_window=initialize_window, change_menu=change_menu, optimization=optimization)
        self.__inertia_coefficient_label: tk.Label = tk.Label(text=f"Inertia: {optimization.swarm.get_inertia_coefficient()}", **self.__widget_parameters)
        self.__iterations_label: tk.Label = tk.Label(text=f"N. iterations: {optimization.iterations}", **self.__widget_parameters)

    def display(self, parent_width: int) -> None:
        # labels_list: list[tk.Label] = [attr for attr in dir(self) if attr[-5:] == "label"]
        # * There could be a better way of doing it using a list
        
        # * Sets the optimization frame in the parent frame
        for col in range(0, 5):
            self.root.columnconfigure(col, weight=1)
        for row in range(0, 3):
            self.root.rowconfigure(row, weight=1)
        self.__name_label.grid(row=0, column=0, sticky="nsew")
        self.__function_label.grid(row=1, column=0, sticky="nsew")
        self.__dimensions_label.grid(row=2, column=0, sticky="nsew")
        self.__minima_indicator_label.grid(row=0, column=1, sticky="nsew")
        self.__cognitive_coefficient_label.grid(row=1, column=1, sticky="nsew")
        self.__num_particles_label.grid(row=2, column=1, sticky="nsew")
        self.__minima_value_label.grid(row=0, column=2, sticky="nsew")
        self.__social_coefficient_label.grid(row=1, column=2, sticky="nsew")
        self.view_button.grid(row=0, column=3, sticky="nsew")
        self.__inertia_coefficient_label.grid(row=1, column=3, sticky="nsew")
        self.__iterations_label.grid(row=2, column=3, sticky="nsew")
        self.root.place(x=((parent_width - self.__width - self.__scrollbar_width)//2), y=self.__index*(self.__height + self.__separation) + self.__separation, width=self.__width, height=self.__height)
        # ? Maybe raise an error if the x- or y- coordinates do not make sense


if __name__ == "__main__":
    optim = OptimizationFrame(tk.Tk(), Optimization(1), 50, 30)
    optim.display()




import tkinter as tk
from tkinter import font

from psopackage.graphics.colors import Color
from psopackage.graphics.fonts import FontName

class CreateInput:
    def __init__(self, parent_frame, default_value: str, text: str, width: int):
        self.__label: tk.Label = tk.Label(parent_frame, text=text, font=font.Font(family=FontName.label, size=12), bg=Color.create_label_bg, fg=Color.create_label_fg)
        self.input_value: tk.StringVar = tk.StringVar()
        self.entry: tk.Entry = tk.Entry(parent_frame, width=width, justify="center", bg=Color.create_entry_bg, fg=Color.create_entry_fg, textvariable=self.input_value)
        self.__default_value: str = default_value
        self.input_value.set(default_value)
        self.entry.bind("<FocusIn>", self.__select_text)

    def __select_text(self, e) -> None:
        self.entry.select_range(0, "end")
        self.entry.icursor("end")

    def grid(self, label_row: int, column: int, sticky: str):
        self.__label.grid(row=label_row, column=column, sticky=sticky, padx=(10, 0), pady=20)
        self.entry.grid(row=label_row, column=column + 1, sticky=sticky, padx=(0, 10), pady=20)

    def get_input(self) -> str:
        return self.input_value.get()


"""
This module defines the Particle class, which represents a particle in a swarm in the context of the Particle Swarm Optimization (PSO) algorithm.

## Classes
- Particle: Represents a particle in a swarm in the context of the PSO algorithm.

### Methods
- initialize_randomly(bound=10): Initializes the position, velocity, heuristic, and pbest of the particle randomly.
- _update_pbest(): Updates the best position of the particle (__pbest) comparing the heuristic value.

#### Getters and setters
- get_heuristic() -> Heuristic
- get_pbest() -> Position
- get_position() -> Position
- get_velocity() -> Velocity
- set_heuristic(heuristic: Vector) -> None
- set_pbest(pbest: Position) -> None
- set_position(position: Position) -> None
- set_velocity(velocity: Velocity) -> None
"""

import numpy as np

from psopackage.vector.base_vector import Vector
from psopackage.vector.heuristic import Heuristic
from psopackage.vector.position import Position
from psopackage.vector.velocity import Velocity

class Particle:
    """
    Particle class represents a particle in a swarm in the context of 
    the Particle Swarm Optimization (PSO) algorithm.

    ## Attributes
    
    ### Private
    - heuristic : Heuristic
        Fitness or heuristic value of the particle.
    - pbest : Position
        Position where the best heuristic value was found.
    - position : Position
        Current position of the particle.
    - velocity : Velocity
        Current velocity of the particle.
    
    ### Public
    - color : dict
        Color of the particle's heuristic.
    - has_gbest : bool
        If the particle has a global best position.

    ## Methods
    - initialize_randomly(bound=10)
        Initializes the position, velocity, heuristic and pbest of the particle randomly.

    ### Getters and setters

    - get_heuristic() -> Vector
        Returns the fitness or heuristic value of the particle.
    - get_pbest() -> Position
        Returns the position where the best heuristic value was found.
    - get_position() -> Position
        Returns the current position of the particle.
    - get_velocity() -> Velocity
        Returns the current velocity of the particle.
    - set_heuristic(heuristic: Vector) -> None
        Sets the fitness or heuristic value of the particle.
    - set_pbest(pbest: Position) -> None
        Sets the position where the best heuristic value was found.
    - set_position(position: Position) -> None
        Sets the current position of the particle.
    - set_velocity(velocity: Velocity) -> None
        Sets the current velocity of the particle.
    """
    # ! Update documentation

    def __init__(self, heuristic_f: callable, index: int, has_gbest: bool, cognitive_coefficient: float = 2.0, dimensions: int = 3, inertia_coefficient: float = 1.0, social_coefficient: float = 2.0) -> None:
        self.__cognitive_coefficient: float = cognitive_coefficient
        self.__index: int = index
        self.__inertia_coefficient: float = inertia_coefficient
        self.__social_coefficient: float = social_coefficient
        self.__position: Position = Position(dimensions-1)
        self.__pbest: Position = Position(dimensions-1)
        self.__heuristic: Heuristic = Heuristic(heuristic_f, dimensions)
        self.__heuristic._update(self.get_pbest())
        self.__velocity: Velocity = Velocity(dimensions-1)
        self.color : dict = {"r": 0, "g": 0, "b": 0}
        self.has_gbest = has_gbest

    def __repr__(self) -> str:
        return f"Particle at {self.__position.get_coordinates()} with pbest at {self.__pbest.get_coordinates()}, velocity {self.__velocity.get_coordinates()} and a heuristic value of {self.__heuristic.get_coordinates()}.\n"
    
    def initialize_randomly(self, bound: int = 10) -> None:
        """Initializes the (__)position, (__)velocity, (__)heuristic and (__)pbest of the particle randomly.
        
        ## Parameters
        bound : int
            The upper bound for the random initialization. Default is 10.
        """
        self.__position.initialize_randomly(bound, self.__position.get_dimensions())
        self.__velocity.initialize_randomly(bound / 5, self.__position.get_dimensions())
        self.__position._update(self.__velocity)
        self.__heuristic._update(self.__position)
        self.__pbest.set_coordinates(self.__position.get_coordinates().copy())

    def _update_velocity(self, gbest: Position) -> None:
        # ! There must be something wrong with this method: the operation is not being performed correctly.
        # * It may be correct, just the parameters not the ideal ones.
        """
        Updates the velocity vector based on the particle swarm optimization
        formula:
        v(i + 1) = w * v(i) + c1 * r1 * (pbest - x(i)) + c2 * r2 * (gbest - x(i))

        ### Notes
        - The `position`, `pbest`, and `gbest` constants should have the same shape as the velocity vector.
        - The `w` constant (inertia coefficient) ranges between 0 and 1.
        - The `c1` (cognitive coefficient) and `c2` (social coefficient) constants range between 1 and 3.
        - The `r1` and `r2` values are random numbers between 0 and 1.
        """

        x_i: np.ndarray = self.__position.get_coordinates()
        pbest: np.ndarray = self.__pbest.get_coordinates()
        initial_velocity: np.ndarray = self.__velocity.get_coordinates()
        r1: float = np.random.uniform(0, 1)
        r2: float = np.random.uniform(0, 1)
        w: float = self.__inertia_coefficient
        c1: float = self.__cognitive_coefficient
        c2: float = self.__social_coefficient
        final_velocity: np.ndarray = (w * initial_velocity) + (c1 * r1 * (pbest - x_i)) + (c2 * r2 * (gbest.get_coordinates() - x_i))
        final = np.clip(final_velocity, -5, 5)
        self.__velocity.set_coordinates(final)

    def _update_pbest(self) -> None:
        """Updates the best position of the particle (__pbest) comparing the heuristic
        value (last coordinate of the heuristic vector) of the current position 
        (__position) with the heuristic value of the best position found up to
        the i-th iteration (__pbest)."""
        heuristic_f = self.__heuristic.get_heuristic_f()
        if heuristic_f(self.__position) < heuristic_f(self.__pbest):
            self.__pbest.set_coordinates(self.__position.get_coordinates().copy())
    
    # * Getters and setters

    def get_heuristic(self) -> Heuristic:
        return self.__heuristic
    
    def get_index(self) -> int:
        return self.__index
    
    def get_pbest(self) -> Position:
        return self.__pbest
    
    def get_position(self) -> Position:
        return self.__position
    
    def get_velocity(self) -> Velocity:
        return self.__velocity
    
    def set_heuristic(self, heuristic: Vector) -> None:
        self.__heuristic = heuristic
    
    def set_pbest(self, pbest: Position) -> None:
        self.__pbest = pbest
    
    def set_position(self, position: Position) -> None:
        self.__position = position
    
    def set_velocity(self, velocity: Velocity) -> None:
        self.__velocity = velocity
        

if __name__ ==  "__main__":
    dimensions: int = 4
    p = Particle(0)
    print(p.get_heuristic().get_coordinates())
    print(p.get_pbest().get_coordinates())
    print(p.get_position().get_coordinates())
    print(p.get_velocity().get_coordinates())



"""
A module to run the Particle Swarm Optimization (PSO) algorithm.

## Classes
Main: Main class to run the Particle Swarm Optimization (PSO) algorithm.

### Methods
- graph_heuristic() -> None
- graph_particles() -> None
- heuristic(position: Position) -> float
- optimize() -> None

#### Getters
- get_cognitive_coefficient() -> float
- get_dimensions() -> int
- get_inertia_coefficient() -> float
- get_iterations() -> int
- get_particle_amount() -> int
- get_social_coefficient() -> float
- get_swarm() -> ParticleSwarm
"""
import math
from matplotlib.pyplot import Figure

import numpy as np
import pandas as pd

from psopackage.swarm.particleSwarm import ParticleSwarm
from psopackage.vector.position import Position
from psopackage.database.data import Data

class Optimization:
    def __init__(self, index: int, function_fig: Figure, function_choice: str = "Sphere", data: Data = Data("test"), cognitive_coefficient: float = 2.05, inertia_coefficient: float = 0.7, social_coefficient: float = 2.05, particle_amount: int = 10, dimensions: int = 3, iterations: int = 20) -> None:
        self.__data: Data = data
        self.__iterations: int = iterations
        # * So it doesn't create two particle swarms with different dimensions
        self.__function_choice: str = function_choice
        self.__swarm: ParticleSwarm = ParticleSwarm(self.heuristic(function_choice), inertia_coefficient, cognitive_coefficient, social_coefficient, dimensions, particle_amount)
        self.__index: int = index
        self._dimensions: int = dimensions
        self.function_fig: Figure = function_fig
        self.__optimization_df: pd.DataFrame = None
        self.bound: float = 0

    @property
    def optimization_df(self) -> pd.DataFrame:
        return self.__optimization_df

    def sphere_f(self, position: Position):
        return np.sum(np.square(position.get_coordinates()))
    
    def booth_f(self, position: Position):
        x = position.get_coordinates()[0]
        y = position.get_coordinates()[1]
        return (x + 2*y - 7)**2 + (2*x + y - 5)**2
    
    def goldstein_price_f(self, position: Position):
        x = position.get_coordinates()[0]
        y = position.get_coordinates()[1]
        return((1 + (x+y+1)**2 * (19 - 14 * x + 3 * x**2 - 14 * y + 6*x*y + 3*y**2)) * (30 + (2*x - 3*y)**2 * (18 - 32 * x + 12 * x**2 + 48 * y - 36*x*y + 27 * y**2)))

    def rastrigin_f(self, position: Position):
        return 20 + np.sum(np.square(position.get_coordinates()) - 10*np.cos(2 * np.pi *position.get_coordinates()))
    
    def heuristic(self, selection: str) -> callable:
        """Heuristic function to be optimized."""
        
        if selection == "Sphere":
            self.bound = 2
            return self.sphere_f
        
        if selection == "Booth": #(1,3)
            self.bound = 10
            return self.booth_f
        
        if selection == "Goldstein-Price": #(0,-1)
            self.bound = 2
            return self.goldstein_price_f
        
        elif selection == "Rastrigin":
            self.bound = 5.12
            return self.rastrigin_f
                                
        else:
            raise NotImplementedError(f"Function {selection} not implemented")
        
    def optimize(self) -> None:
        """Optimizes the heuristic function using the PSO algorithm."""
        self.heuristic(self.__function_choice) # * So the bounds are initialized correctly.
        swarm = self.__swarm
        swarm._initialize_particles_randomly(bound=self.bound)
        swarm.update_gbest()
        swarm_gbest_index: list[int] = []
        optimization_df: pd.DataFrame = pd.DataFrame(columns = ["Heuristic",
                                            "Position", "Velocity", "Pbest"])
        # * A df of nan's is created to separate optimizations more evidently when passing them to the database and when showing them in the GUI
        nan_df = pd.DataFrame(([np.nan] * 4), index = ["Heuristic", "Position",
                                                       "Velocity", "Pbest"]).T

        for iteration_num in range(self.__iterations + 1):
            iteration_data: dict = {"Heuristic": [], "Position": [],
                                    "Velocity": [], "Pbest": []}
            for particle in swarm.get_particles():
                if iteration_num > 0: 
                    # * To record the initial states of the particles before optimizing them
                    # * Each iteration the particles update their velocity and position.
                    particle._update_velocity(swarm.get_gbest())
                    particle.get_position()._update(particle.get_velocity())
                    current_coordinates: np.ndarray = particle.get_position().get_coordinates().copy()
                    new_coordinates: np.ndarray = np.clip(current_coordinates, -self.bound, self.bound)
                    particle.get_position().set_coordinates(new_coordinates)
                    particle.get_heuristic()._update(particle.get_position())
                    particle.get_velocity().set_coordinates(np.clip(particle.get_position().get_coordinates(), -5, 5))                    
                    particle._update_pbest()
                    

                # * Append the data of each particle after a certain iteration
                # * to a temporary dictionary
                # ? Should the np.ndarrays be copies?
                iteration_data["Heuristic"].append(np.round(particle.
                    get_heuristic().get_coordinates().copy(), 2))
                iteration_data["Position"].append(np.round(particle.
                    get_position().get_coordinates().copy(), 2))
                iteration_data["Velocity"].append(np.round(particle.
                    get_velocity().get_coordinates().copy(), 2))
                iteration_data["Pbest"].append(np.round(particle.
                    get_pbest().get_coordinates().copy(), 2))

            # * Append the last iteration's data and the index of the particle
            # * with the best heuristic to the database.
            swarm.update_gbest()
            # print(swarm.get_gbest())
            optimization_df = pd.concat([optimization_df, pd.DataFrame(iteration_data)])
            # TODO: Check the logic behind the if.
            # * Seems to be working fine
            if iteration_num != self.__iterations:
                optimization_df = pd.concat([optimization_df, nan_df])
            print(f"Global best: {swarm.get_gbest()}, Heuristic value:{swarm.get_heuristic()(swarm.get_gbest())}\n")

        # * Append the indexes of the particles with the best heuristic to the
        # * database and create a spreadsheet with the optimization results.
        self.__data.append_gbest_indexes(swarm_gbest_index)
        self.__data.append_optimization(optimization_df)
        self.__optimization_df = optimization_df
    
    # Applying decorators to the project
    @property
    def function_choice(self) -> str:
        return self.__function_choice

    @property
    def dimensions(self) -> int:
        return self._dimensions
    
    @property
    def index(self) -> int:
        return self.__index
    
    @property
    def iterations(self) -> int:
        return self.__iterations

    @property
    def swarm(self) -> ParticleSwarm:
        return self.__swarm
    
if __name__ == "__main__":
    data = Data(excel_file_name="session1_results")
    main = Optimization(0, data=data, cognitive_coefficient=1.5, inertia_coefficient=0.4, social_coefficient=1.5, particle_amount=20, dimensions=3, iterations=100) # ! CHECK: Minimum dimension
    main.optimize()
    print(main.get_swarm().get_gbest())



import tkinter as tk
from tkinter import font

from psopackage.graphics.mainMenu.bottomFrame import BottomFrame
from psopackage.graphics.mainMenu.optionsFrame import OptionsFrame
from psopackage.graphics.fonts import FontName
from psopackage.graphics.colors import Color

class MainMenu():
    """ 
    Creates the main menu of the GUI.
    """
    def __init__(self, master_frame: tk.Frame, initialize_window: callable, change_menu: callable, program_version: str, window_width: int=250, window_height: int=250):
        self._parent_frame: tk.Frame = master_frame # ! Check if this is the right root
        self.root: tk.Frame = tk.Frame(master_frame, bg=Color.test2_bg)
        self.__initialize_window: callable = initialize_window
        self._title: tk.Label = tk.Label(self.root, text="PSO manager",
            bg=Color.optim_label_bg, font=font.Font(family=FontName.title, size=15), wraplength=450,
            anchor="center")
        self._window_width: int = window_width
        self._window_height: int = window_height
        self._title_height: int = 40
        self._bottom_frame_height: int = 30
        self.__bottom_frame: BottomFrame = BottomFrame(master_frame, self.root, self._window_width, self._window_height, self._title_height, self._bottom_frame_height, program_version)
        self.__options_frame: OptionsFrame = OptionsFrame(self.root, change_menu, self._window_width, self._window_height, self._title_height, self._bottom_frame_height)

    # * The baseMenu idea was discarded to avoid a circular import with GUI.
            
    def display(self):
        """
        """
        self.__initialize_window(width=250, height=250)
        self._title.place(x=0, y=0, width=self._window_width,
            height=self._title_height)
        self.__bottom_frame.display()
        self.__options_frame.display()
        self.root.place(x=0, y=0, width=250, height=250)

    def forget(self):
        self.root.place_forget()
        self.__bottom_frame.place_forget()
        self.__options_frame.place_forget()
        # TODO: Add padding to popUpframe
        
    


"""
A module to represent a vector as a data type with added functionality.
Contains the class Vector and its related attributes and methods.
It is intended to be an abstract class to be inherited by other classes.

## Classes
Vector: A class to represent a vector in a n-dimensional space and
    facilitate its graphical representation.

### Methods
- initialize_randomly(bound: float = 10): Initialize the coordinates of a
vector randomly within
- _update(): Abstract method.

#### Getters and setters
- get_coordinates() -> np.ndarray
- get_dimensions() -> int
- set_coordinates(coordinates: np.ndarray)
- set_dimensions(dimensions: int)

"""

from typing import Any

import numpy as np

class Vector:
    """A class to represent a vector in a n-dimensional space and
    facilitate its graphical representation.

    ## Parameters
    - dimensions : int
        The number of dimensions of the vector. Default is 3.
    
    ## Attributes
    - _coordinates : np.ndarray
        An array with the coordinates of the vector.
    - _dimensions : int
        The number of dimensions of the vector.
    
    ## Methods
    - __repr__() -> str
        Overwrites the __repr__ method to return a string with the
        coordinates of the vector.
    - initialize_randomly(bound: float = 10)
        Initialize the coordinates of a vector randomly within 
        the interval [-bound, bound].
    
    Getters and setters
    - get_color() -> dict
    - get_coordinates() -> np.ndarray
    - get_dimensions() -> int
    - set_color(red: int, green: int, blue: int, alpha: int = 255)
    - set_coordinates(coordinates: np.ndarray)
    - set_dimensions(dimensions: int) 
        
    """
    
    def __init__(self, dimensions = 3) -> None:
        self._coordinates: np.ndarray = np.zeros(dimensions)
        self._dimensions: int = dimensions

    def __repr__(self) -> str:
        return f"Vector with coordinates {self.get_coordinates()}."
    
    def initialize_randomly(self, bound: float = 10, dimensions: int = 2) -> None:
        """Initialize the coordinates of a vector randomly within 
        the interval [-bound, bound].

        ## Parameters
        bound (float): 
            The maximum absolute value of the coordinates in all of its dimmensions.
        """
        self._coordinates = np.random.uniform(low=-bound, high=np.nextafter(bound, bound + 1), size=dimensions) 
        self._coordinates = np.clip(self._coordinates, -bound, bound)
    
    # def __lt__(self, other) -> bool:
    #     if not isinstance(other, Vector):
    #         raise TypeError("The object to compare must be a Vector. The provided object is of type {}.".format(type(other)))
    #     print(other.get_dimensions(), self._dimensions)
    #     return self._coordinates[self._dimensions - 1] < other._coordinates[self._dimensions - 1]

    # * Getters and setters
    
    def get_coordinates(self) -> np.ndarray:
        return self._coordinates
    
    def get_dimensions(self) -> int:
        return self._dimensions

    def set_coordinates(self, coordinates: np.ndarray) -> None:
        self._coordinates = coordinates
    
    def set_dimensions(self, dimensions: int) -> None:
        self._dimensions = dimensions
    
    # It is another option to access the attributes of the class, but it may
    # not be what we want to do since it violates private access
    # def __getattribute__(self, name: str) -> Any:
    #     attributes: dict = {
    #         "color": super().__getattribute__("_Vector_color"), 
    #         "coordinates": super().__getattribute__("_Vector_coordinates"), 
    #         "dimensions": super().__getattribute__("_Vector_dimensions")
    #     }
    #     if name in attributes:
    #         return attributes[name]
    #     else:
    #         return super().__getattribute__(name)
    
    # def __setattr__(self, name: str, value: Any) -> None:
    #     attributes: dict = {
    #         "color": "_Vector_color", 
    #         "coordinates": "_Vector_coordinates", 
    #         "dimensions": "_Vector_dimensions"
    #     }
    #     if name in attributes:
    #         super().__setattr__(attributes[name], value)
    #     else:
    #         super().__setattr__(name, value)
        
if __name__ == "__main__":
    vect1: Vector = Vector(3)
    print("Dimensions (set to 3):", vect1.get_dimensions())

    vect1.initialize_randomly(np.random.randint(1, 11))
    print("Coordinates randomly initialized:", vect1.get_coordinates())
    vect1.set_coordinates(np.array([1, 2, 3]))
    print("Coordinates set to [1, 2, 3]:", vect1.get_coordinates())

    red: int = np.random.randint(0, 256)
    green: int = np.random.randint(0, 256)
    blue: int = np.random.randint(0, 256)
    vect1.set_color(red, green, blue)
    print("Set random colors with default alpha:", vect1.get_color())


import tkinter as tk
from tkinter import font

from psopackage.graphics.colors import Color
from psopackage.graphics.fonts import FontName

class FunctionChoiceMenu():
    """ 
    Class that manages teh function to get chosen from a dropdown menu.
    """
    def __init__(self, parent_frame: tk.Frame, text: str, options: list[str], display_graph: callable) -> None:
        self.__options: list[str] = options
        self.__display_graph: callable = display_graph
        self.__choice: tk.StringVar = tk.StringVar()
        self.__choice.set(options[0])
        self.__choice.trace_add("write", self.__trigger_graph_change)
        self.__dropdown_menu: tk.OptionMenu = tk.OptionMenu(parent_frame, self.__choice, *options)
        self.__label: tk.Label = tk.Label(parent_frame, text=text, bg=Color.create_label_bg, fg=Color.create_label_fg, font=font.Font(family=FontName.label, size=12))

    def __trigger_graph_change(self, *args) -> None:
        self.__display_graph(self.__choice.get())

    def grid(self, label_row: int, column: int, sticky: str) -> None:
        self.__choice.set(self.__options[0])
        self.__label.grid(row=label_row, column=column, sticky=sticky, padx=(10, 0), pady=20)
        self.__dropdown_menu.grid(row=label_row, column=column + 1, sticky=sticky, padx=(0, 10), pady=20) # ! Padding seems to have a limited effect in an OptionMenu

    def get_choice(self) -> str:
        return self.__choice.get()


import tkinter as tk

from psopackage.graphics.colors import Color

class BackButton(tk.Button):
    """ 
    A class that manages the back button in the GUI. Inherits from tk.Button.
    """
    def __init__(self, menu_frame, image: tk.PhotoImage, active_image: tk.PhotoImage, width: int, height: int,  change_menu: callable, change_menu_args: dict = {"menu_name": "main"}, cbg: str = Color.back_button_cbg, args: dict = {
        "bg": Color.back_button_bg,
        "activebackground": Color.back_button_abg,
        "relief": "flat", 
        "cursor": "hand2",
        "highlightcolor": Color.back_button_hcolor,
        "highlightthickness": 1
        }):
        super().__init__(menu_frame, image=image, **args)
        self.__image: tk.PhotoImage = image
        self.__active_image: tk.PhotoImage = active_image
        self.__width: int = width
        self.__height: int = height
        self.__change_menu: callable = change_menu
        self.__change_menu_args: dict = change_menu_args

    def __enter(self, e) -> None:
        self.config(bg=Color.back_button_abg, activebackground=Color.back_button_abg, image=self.__active_image)

    def __leave(self, e) -> None:
        self.config(bg=Color.back_button_bg, image=self.__image)

    def __click(self, e) -> None:
        self.config(bg=Color.back_button_cbg, activebackground=Color.back_button_cbg, image=self.__image)

    def __release(self, e) -> None:
        # * Goes back to the main menu
        self.__change_menu(**self.__change_menu_args)

    # ! Consider binding to events in the initializer so that every time a widget is displayed it doesn't need to bind events again.
    # * To keep the code consistent, might be better to leave it there, because BottomButton objects do need it to be there

    def __bind_to_events(self) -> None:
        self.bind("<Enter>", self.__enter)
        self.bind("<Leave>", self.__leave)
        self.bind("<Button-1>", self.__click)
        self.bind("<ButtonRelease-1>", self.__release)
        self.bind("<KeyRelease-Return>", self.__release)

    def display(self, x: int, y: int) -> None:
        self.__bind_to_events()
        self.place(x=x, y=y, width=self.__width, height=self.__height)


class Color:

    # * TESTS
    test1_bg: str = "#ff0000"
    test2_bg: str = "#00ff00"
    test3_bg: str = "#0000ff"
    # ? Would it be better to do it with a dictionary?
    # * WINDOWS AND FRAMES

    bottom_menu_bg: str = "#0d42a3"
    help_frame_bg: str = "#a1adc4"
    info_frame_bg: str = "#a1adc4"
    window_bg: str = "#f0f0f0"
    goodbye_frame_bg: str = "#1b1b1b"

    # ! Consider implementing hbg in all buttons

        # * Select menu
    select_inner_frame_bg: str = "#afcff8"
    select_inner_frame_fg: str = "#000000"
    select_inner_button_bg: str = "#f0f0f0"

    # * Create optimization menu
    create_entry_bg: str = "#d3d3d3"
    create_entry_fg: str = "#000000"

    create_label_bg: str = "#d3d3d3"
    create_label_fg: str = "#000000"

    create_button_bg: str = "#95d8ec"
    create_button_fg: str = "#000000"
    create_button_abg: str = "#d3d3d3"
    create_button_afg: str = "#000000"
    create_button_hbg: str = "#d3d3d3"
    create_button_hcolor: str = "#000000"
    create_button_cbg: str = ""

    # * BUTTONS, LABELS AND TEXT

    # * Background
    # * Main menu
    bottom_button_bg: str = "#0d42a3"
    hide_button_bg: str = "#a3490d"
    optim_button_bg: str = "#89e379"

    bottom_label_bg: str = info_frame_bg
    optim_label_bg: str = "#d3d3d3"
    goodbye_label_bg: str = goodbye_frame_bg

    goodbye_text_bg: str = goodbye_frame_bg

    # * Select menu
    select_button: str = "#f0f0f0"
    select_label_optim_bg: str = "#d3d3d3"
    select_label_no_optim_bg: str = "#d1d1d1"
    select_title_bg: str = "#d3d3d3"
    back_button_bg = select_title_bg
    preview_button_bg: str = select_label_optim_bg

    # * Foreground (font color)
    # * Main menu
    bottom_button_fg: str = "#dfdfdf"
    optim_button_fg: str = "#000000"
    hide_button_fg: str = "#dfdfdf"

    bottom_label_fg: str = "#000000"
    goodbye_text_fg: str = "#dfdfdf"

    # * Select menu
    select_label_optim_fg: str = "#000000"
    select_label_no_optim_fg: str = "#000000"
    select_title_fg: str = "#000000"
    
    # * Active background
    # * Main menu
    bottom_button_abg: str = "#225fd1"
    optim_button_abg: str = "#123f0a" # ! Consider changing optim (gets confused in select menu)
    hide_button_abg: str = "#d25e0a"

    # * Select menu
    back_button_abg: str = "#f0f0f0"
    preview_button_abg: str = "#f7f7f7"

    # * Active foreground
    # * Main menu
    bottom_button_afg: str = "#ffffff"
    optim_button_afg: str = "#ffffff"
    hide_button_afg: str = "#000000"

    # * Select menu
    back_button_afg: str = "#000000"

    # * Highlight background (focused)
    # * Main menu
    bottom_button_hbg: str = bottom_button_bg
    hide_button_hbg: str = "#a23d0a"
    optim_button_hbg: str = optim_button_bg

    # * Select menu
    back_button_hbg: str = "#922929"
    preview_button_hbg: str = preview_button_abg

    # * Highlight color (focused)
    # * Main menu
    bottom_button_hcolor: str = "#ffffff"
    hide_button_hcolor: str = "#000000"
    optim_button_hcolor: str = optim_button_abg

    # * Select menu
    back_button_hcolor: str = "#000000"
    preview_button_hcolor: str = preview_button_abg


    # * Clicked background
    # * Main menu
    bottom_button_cbg: str = "#033186" # Todo: Consider changing it
    optim_button_cbg: str = "#012422"
    hide_button_cbg: str = "#a23d0a"

    # * Select menu
    back_button_cbg: str = back_button_bg
    preview_button_cbg: str = "#e0e0e0"

    # * Clicked foreground
    # * Main menu
    optim_button_cfg: str = optim_button_afg
    hide_button_cfg: str = hide_button_fg

    # * Select menu
    back_button_cfg: str = "#000000"

    # * SCROLLBAR
    # * Main menu
    bottom_scrollbar_bg: str = "#2b2a2a"
    bottom_scrollbar_trough: str = bottom_label_bg
    bottom_scrollbar_abg: str = "#0f0f0f"
    bottom_scrollbar_hbg: str = bottom_label_bg
    bottom_scrollbar_hcolor: str = "#000000"
    
    # * Data colors
    siders: str = "ffffff"
    common_side: str = "000000"
    title_cell: str = "f0f0f0"
    grad_fill_start: str = "085063"
    grad_fill_stop: str = "d6e416"
    cell_fill: str = "042d53"
    cell2_fill: str = "dfdfdf"


"""
This module defines the ParticleSwarm class, which represents multiple Particle objects in a swarm in the context of the Particle Swarm Optimization (PSO) algorithm.

## Classes
- ParticleSwarm: Represents a particle swarm optimization algorithm.

### Attributes
- __inertia_coefficient: float - The inertia coefficient used in the particle update equation.
- __cognitive_coefficient: float - The cognitive coefficient used in the particle update equation.
- __social_coefficient: float - The social coefficient used in the particle update equation.
- __particle_amount: int - The number of particles in the swarm.
- __particles: list[Particle] - The list of particles in the swarm.
- __gbest: Position - The global best position found by the swarm.
- _heuristic_f: callable - The heuristic function to be optimized.

### Methods
- __init__(inertia_coefficient: float = 1, cognitive_coefficient: float = 2, social_coefficient: float = 2, dimensions: int = 3, particle_amount: int = 10, heuristic: callable = default_heuristic) -> None: Initializes the particle swarm with the given parameters.
- __repr__() -> str: Returns a string representation of the particle swarm.
- _initialize_particles_randomly(bound: float = 10) -> None: Initializes the positions and velocities of particles randomly.
- update_gbest() -> None: Updates the global best position found by the swarm.

### Getters
- get_inertia_coefficient() -> float: Returns the inertia coefficient.
- get_cognitive_coefficient() -> float: Returns the cognitive coefficient.
- get_social_coefficient() -> float: Returns the social coefficient.
- get_particle_amount() -> int: Returns the number of particles in the swarm.
- get_particles() -> list[Particle]: Returns the list of particles in the swarm.
- get_gbest() -> Position: Returns the global best position found by the swarm.
- get_heuristic() -> callable: Returns the heuristic function to be optimized.
"""
import numpy as np

from psopackage.swarm.particle import Particle
from psopackage.vector.position import Position

class ParticleSwarm:
    """
    Represents a particle swarm optimization algorithm.

    ## Parameters
    - inertia_coefficient : float, optional
        The inertia coefficient used in the particle update equation. Default is 1.
    - cognitive_coefficient : float, optional
        The cognitive coefficient used in the particle update equation. Default is 2.
    - social_coefficient : float, optional
        The social coefficient used in the particle update equation. Default is 2.
    - dimensions : int, optional
        The number of dimensions in the search space. Default is 3.
    - particle_amount : int, optional
        The number of particles in the swarm. Default is 10.
    - heuristic : callable, optional
        The heuristic function to be optimized. Default is the default_heuristic function
        imported from the heuristic module.

    ## Attributes
    - __inertia_coefficient : float
        The inertia coefficient used in the particle update equation.
    - __cognitive_coefficient : float
        The cognitive coefficient used in the particle update equation.
    - __social_coefficient : float
        The social coefficient used in the particle update equation.
    - __particle_amount : int
        The number of particles in the swarm.
    - __particles : list[Particle]
        The list of particles in the swarm.
    - __gbest : Vector
        The global best position found by the swarm.
    - _heuristic_f : callable
        The heuristic function to be optimized.

    ## Methods
    - __repr__() 
        Returns a string representation of the particle swarm (overridden).
    - _initialize_particles_randomly(bound=10)
        Initializes the positions and velocities of particles randomly.
    - update_gbest()
        Updates the global best position found by the swarm.

    ### Getters
    - get_inertia_coefficient() -> float
        Returns the inertia coefficient.
    - get_cognitive_coefficient() -> float
        Returns the cognitive coefficient.
    - get_social_coefficient() -> float
        Returns the social coefficient.
    - get_particle_amount() -> int
        Returns the number of particles in the swarm.
    - get_particles() -> list[Particle]
        Returns the list of particles in the swarm.
    - get_gbest() -> Vector
        Returns the global best position found by the swarm.
    - get_heuristic() -> callable
        Returns the heuristic function to be optimized.
    """

    def __init__(self, heuristic_f: callable, inertia_coefficient: float = 1, cognitive_coefficient: float = 2, social_coefficient: float = 2, dimensions: int = 3, particle_amount: int = 10) -> None:
        self.__inertia_coefficient: float = inertia_coefficient
        self.__cognitive_coefficient: float = cognitive_coefficient
        self.__social_coefficient: float = social_coefficient
        try:
            self.__particle_amount: int = int(particle_amount)
            if self.__particle_amount < 1:
                raise ValueError("The amount of particles must be greater than zero.") # ! Maybe include a TypeError to deal with floats
            if dimensions <= 1:
                raise IndexError("The number of dimensions must be greater than one.")
        except ValueError as e:
            print(e)
            print("Amount defaulted to 10.")
            self.__particle_amount = 10
        except IndexError as e:
            print(e)
            print("Dimensions defaulted to 3.")
            dimensions = 3
        # TODO: Except TypeError (double)
        # ? Should the following line be inside a finally block?
        finally:
            self.__particles: list[Particle] = [
                Particle(heuristic_f=heuristic_f, index=p, has_gbest=False, cognitive_coefficient=cognitive_coefficient,
                dimensions=dimensions, 
                inertia_coefficient=inertia_coefficient, 
                social_coefficient=social_coefficient) 
                for p in range(self.__particle_amount)
                ] # ! Test change of Particle's constructor
            self.__gbest: Position = Position(dimensions - 1)
            self._heuristic_f: callable = heuristic_f
        self.__bound: int = None
    
    def __repr__(self) -> str:
        return f"Particle swarm with {self.get_particle_amount()} particles, cognitive coefficient {self.get_cognitive_coefficient()}, inertia coefficient {self.get_inertia_coefficient()}, social coefficient {self.get_social_coefficient()} and global best position {self.get_gbest().get_coordinates()}."

    def _initialize_particles_randomly(self, bound: float = 10) -> None:
        """Calls the initialize_randomly method of each particle in the
        swarm and then updates the global best position (__gbest).
        
        ## Parameters
        bound : float
            The bound used to delimit the values of the particles' positions. Default is 10.
        """
        self.__bound = bound
        for particle in self.__particles:
            particle.initialize_randomly(bound)
        self.__gbest.set_coordinates(self.__particles[0].get_position().get_coordinates().copy())
        # print("BBBB", self.__gbest.get_coordinates())
        self.update_gbest()

    def update_gbest(self) -> None:
        """Compares each particles' position (accessed through get_position()) 
        with the global best position (__gbest)"""
        dimensions: int = self.__particles[0].get_position().get_dimensions()
        gbest_index: int = 0
        for particle in self.__particles:
            if particle.get_heuristic().get_coordinates()[dimensions] < self._heuristic_f(self.__gbest):
                # print(f"{particle.get_heuristic().get_coordinates()[dimensions]}  <  {self._heuristic_f(self.__gbest)}")
                self.__gbest.set_coordinates(particle.get_position().get_coordinates().copy())
                gbest_index = particle.get_index()
        self.__particles[gbest_index].has_gbest = True

    # * Getters (setters not necessary for now)

    def get_inertia_coefficient(self) -> float:
        return self.__inertia_coefficient

    def get_cognitive_coefficient(self) -> float:
        return self.__cognitive_coefficient
    
    def get_social_coefficient(self) -> float:
        return self.__social_coefficient
    
    def get_particle_amount(self) -> int:
        return self.__particle_amount
    
    def get_particles(self) -> list[Particle]:
        return self.__particles
    
    def get_gbest(self) -> Position:
        return self.__gbest
    
    def get_heuristic(self) -> callable:
        return self._heuristic_f

if __name__ == "__main__":
    swarm = ParticleSwarm(1, 1, 1, 2, 5)
    swarm._initialize_particles_randomly(5)
    print(swarm.get_particles())
    print(swarm.get_gbest())
    swarm.update_gbest()
    print(swarm.get_gbest())




import tkinter as tk

from psopackage.graphics.colors import Color
from psopackage.graphics.mainMenu.popUpFrame import PopUpFrame

class BottomButton(tk.Button):
    """ 
    This class manages the buttons in the bottom frame of the GUI. Inherits from tk.Button.
    """
    def __init__(self, menu_frame: tk.Frame, parent_frame: tk.Frame, image: tk.PhotoImage, active_image: tk.PhotoImage, pop_up_text: str, name: str, pop_up_frame_height: int, pop_up_frame_width: int, pop_up_frame_y: int, abg: str = Color.bottom_button_abg, cbg: str = Color.bottom_button_cbg, args: dict = {
        "relief": "flat",
        "bg": Color.bottom_button_bg,
        "highlightbackground": Color.bottom_button_hbg,
        "highlightcolor": Color.bottom_button_hcolor,
        "highlightthickness": 1,
        "borderwidth": 0,
        "cursor": "hand2"
        }) -> None:
        super().__init__(parent_frame, image=image, **args)
        self.__image: tk.PhotoImage = image
        self.__active_image: tk.PhotoImage = active_image
        self.__bg: str = args["bg"]
        self.__abg: str = abg
        self.__cbg: str = cbg
        self.pop_up_frame: PopUpFrame = PopUpFrame(menu_frame, name=name, height=pop_up_frame_height, width=pop_up_frame_width, y=pop_up_frame_y, text=pop_up_text)

    def __enter(self, e):
        self.config(bg=self.__abg, activebackground=self.__abg, image=self.__active_image)
    
    def __leave(self, e):
        self.config(bg=self.__bg, image=self.__image)

    def __click(self, e):
        self.config(bg=self.__cbg, activebackground=self.__cbg, image=self.__image)

    def __release(self, event: tk.Event, pop_up_frames: list["PopUpFrame"]):
        # * Implements when the button is released, it will display the pop up frame
        other_frames: list[PopUpFrame] = [frame for frame in pop_up_frames if frame.get_name() != self.pop_up_frame.get_name()]
        if event.type == '3':
            pass
        elif event.type == '5':
            self.config(
                image=self.__active_image,
                activebackground=self.__abg,
                bg=self.__bg
            )
            # ! Could be improved: If we were to manage more than three 
        if self.pop_up_frame.visible == False:
            for frame in other_frames:
                frame.visible = False
                frame.place_forget()
            self.pop_up_frame.visible = True
            self.pop_up_frame.display(other_frames[0])
        
        elif self.pop_up_frame.visible == True:
            self.pop_up_frame.place_forget()
            self.pop_up_frame.visible = False
            for frame in other_frames:
                frame.visible = False
                frame.place_forget()

        else:
            raise NotImplementedError("Frame not implemented in release method")

    def bind_to_events(self, pop_up_frames: dict):
        self.bind("<Enter>", self.__enter)
        self.bind("<Leave>", self.__leave)
        self.bind("<Button-1>", self.__click)
        self.bind("<ButtonRelease-1>", lambda event: self.__release(event, pop_up_frames))
        self.bind("<KeyRelease-Return>", lambda event: self.__release(event, pop_up_frames))

    def display(self, pop_up_frames: list[PopUpFrame], row: int, column: int, sticky: str):
        self.bind_to_events(pop_up_frames)
        self.grid(row=row, column=column, sticky=sticky)



# * This class is supposed to serve as the interaction with all the other
# * classes. It will contain the Data object that will store the results of
# * any optimization done while the program is running.

import os
import re

import toml

from psopackage.database.data import Data
from psopackage.graphics.gui import GUI
from psopackage.optimization import Optimization
class Main:
    def __init__(self) -> None:
        # * The history object is going to store the results of the optimization
        self.optimization_history: list[Optimization] = []
        # * The database object is going to store the results of the optimization
        self.database: Data = Data(excel_file_name=self.determine_xlsx_name())
        print("Data object created")
        # * The GUI object is going to be the main interaction with the user
        self.gui: GUI = GUI(program_version=self.get_version(), data=self.database, optimization_history=self.optimization_history)
    
    def determine_xlsx_name(self) -> str:
        try:
            # * Tries to find the number of sessions. If it fails, it will return session1_results
            session_number = len(os.listdir(path="database/optimization_results")) + 1
            if session_number == 0:
                raise OSError
        except OSError:
            # * In case the directory does not exist or it is empty,
            # * the session is going to be the first one.
            return "session1_results"
        else:
            return f"session{session_number}_results"

    def initialize_optimization(self) -> None:
        # * Appends a new optimization object to the optimization history
        self.__optimizations.append(Optimization(data=self.__history))

    def run_optimization(self) -> None:
        pass

    def get_version(self) -> str:
        # * Gets the version from the pyproject.toml file
        conf_file_path = '../pyproject.toml'
        data = toml.load(conf_file_path)
        version = data.get('tool', {}).get('poetry', {}).get('version', 'Version not found')
        return version

def run():
    main = Main()
    main.gui.run()


if __name__ == "__main__":
    try:
        run()
    except KeyboardInterrupt:
        print()
        print("Exiting the program.")
        exit()





import tkinter as tk
from tkinter import font

from psopackage.graphics.colors import Color
from psopackage.graphics.fonts import FontName

class PopUpFrame(tk.Frame):
    """ 
    Sets up a frame that pops up when a button is clicked.
    """
    def __init__(self, menu_frame: tk.Frame, name: str, height: int, width: int, y: int, text: str):
        super().__init__(menu_frame)
        self.__name: str = name
        self.__text_to_insert: str = text
        text_parameters: dict =  {
            "bg": Color.bottom_label_bg,
            "fg": Color.bottom_label_fg,
            "font": font.Font(family=FontName.label, size=10),
            "wrap": "word",
            "padx": 10,
            "pady": 5,  
            "borderwidth": 0,
            "highlightthickness": 0
        }
        self.__text: tk.Text = tk.Text(self, **text_parameters)
        button_parameters: dict = {
            "bg": Color.hide_button_bg,
            "fg": Color.hide_button_fg,
            "activebackground": Color.hide_button_abg,
            "activeforeground": Color.hide_button_afg,
            "highlightbackground": Color.hide_button_hbg,
            "highlightcolor": Color.hide_button_hcolor,
            "highlightthickness": 1,
            "font": font.Font(family=FontName.label, size=10, weight="bold"),
            "relief": "flat",
            "cursor": "hand2",
            "text": "Hide"
        }
        # self.__text.config(state="normal") # * Not necessary because "normal" is the default behavior of a tk.Text
        self.__text.insert(index="end", chars=self.__text_to_insert)
        self.__text.tag_config(tagName="center", justify="center")
        self.__text.tag_add("center", "1.0", "end") # * Apply center tag to all the characters
        self.__text.config(state="disabled")
        self.__hide_button: tk.Button = tk.Button(self, **button_parameters)
        scrollbar_parameters: dict = {
            "bg": Color.bottom_scrollbar_bg,
            "activebackground": Color.bottom_scrollbar_abg,
            "troughcolor": Color.bottom_scrollbar_trough,
            "highlightbackground": Color.bottom_scrollbar_hbg,
            "highlightcolor": Color.bottom_scrollbar_hcolor,
            "highlightthickness": 1,
            "borderwidth": 0,
            "elementborderwidth": 1,
            "width": 10,
        }
        self.__scrollbar: tk.Scrollbar = tk.Scrollbar(self, scrollbar_parameters)
        self.__width: int = width
        self.__height: int = height
        self.__y: int = y
        self.visible: bool = False

    def __enter_hide_button(self, e) -> None:
        self.__hide_button.config(bg=Color.hide_button_abg, fg=Color.hide_button_afg, activebackground=Color.hide_button_abg, activeforeground=Color.hide_button_afg) # ? Is it necessary to change the abg and the afg?

    def __leave_hide_button(self, e) -> None:
        self.__hide_button.config(bg=Color.hide_button_bg, fg=Color.hide_button_fg)

    def __click_hide_button(self, e) -> None:
        self.__hide_button.config(bg=Color.hide_button_cbg, fg=Color.hide_button_cfg, activebackground=Color.hide_button_cbg, activeforeground=Color.hide_button_cfg)

    def __release_hide_button(self, e, other_frame: "PopUpFrame") -> None:
        self.visible = False
        self.place_forget()
        other_frame.visible = False
        other_frame.place_forget()

    def bind_hide_button(self, other_frame: "PopUpFrame") -> None:
        self.__hide_button.bind("<Enter>", self.__enter_hide_button)
        self.__hide_button.bind("<Leave>", self.__leave_hide_button)
        self.__hide_button.bind("<Button-1>", self.__click_hide_button)
        self.__hide_button.bind("<ButtonRelease-1>", lambda event: self.__release_hide_button(event, other_frame))
        self.__hide_button.bind("<KeyRelease-Return>", lambda event: self.__release_hide_button(event, other_frame))

    def display(self, other_frame: "PopUpFrame"):
        self.bind_hide_button(other_frame)
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=0)
        self.rowconfigure(0, weight=0)
        self.rowconfigure(1, weight=1)
        self.__hide_button.grid(row=0, column=0, columnspan=2, sticky="nsew")
        self.__text.config(yscrollcommand=self.__scrollbar.set)
        self.__scrollbar.config(command=self.__text.yview)
        self.__text.grid(row=1, column=0, sticky="ew")
        self.__text.yview_moveto(0) # * Reset the visible area of the text.
        self.__scrollbar.grid(row=1, column=1, sticky="ns")
        self.place(x=0, y=self.__y, width=self.__width, height=self.__height)

    def get_name(self) -> str:
        return self.__name


import tkinter as tk
from tkinter import font

from psopackage.graphics.colors import Color
from psopackage.graphics.fonts import FontName

class ExitMenu():
    """ 
    Class that manages the exit menu of the GUI.
    """
    def __init__(self, parent_frame: tk.Frame, initialize_window: callable):
        self.__parent_frame: tk.Frame = parent_frame
        self.root: tk.Frame = tk.Frame(parent_frame,
            bg=Color.goodbye_frame_bg)
        self.__initialize_window: callable = initialize_window
        self.__image: tk.PhotoImage = tk.PhotoImage(
            file="graphics/assets/goodbye.png").subsample(2)
        self.__label: tk.Label = tk.Label(self.root, image=self.__image, bg=Color.goodbye_label_bg, borderwidth=0, highlightthickness=0) 
        self.__text: tk.Text = tk.Text(self.root, bg=Color.goodbye_frame_bg, wrap="word", font=font.Font(family=FontName.label, size=25), fg=Color.goodbye_text_fg, background=Color.goodbye_text_bg, borderwidth=0, highlightthickness=0)
        self.__text.insert(index="end", chars="Goodbye!")
        self.__text.tag_config(tagName="center", justify="center")
        self.__text.tag_add("center", "1.0", "end")
        self.__text.config(state="disabled")

    def display(self) -> None:
        self.__initialize_window(width=250, height=250, title="See you later!")
        self.root.rowconfigure(0, weight=0)
        self.root.rowconfigure(1, weight=1)
        self.root.columnconfigure(0, weight=1)
        self.__label.grid(row=0, column=0, pady=(25, 0), sticky="nsew")
        self.__text.grid(row=1, column=0, pady=(25, 25), sticky="nsew")
        # ? A more rigurous way of centering the label and the text could be implemented.
        self.__label.image = self.__image
        self.root.place(x=0, y=0, width=250, height=250)
        self.__parent_frame.after(500, self.__parent_frame.quit)

    def forget(self):
        self.root.forget()

    



from copy import deepcopy
import tkinter as tk
from tkinter import font
from matplotlib.pyplot import Figure, Axes
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import pandas as pd

from psopackage.graphics.colors import Color
from psopackage.graphics.fonts import FontName
from psopackage.graphics.backButton import BackButton
from psopackage.optimization import Optimization

class ViewFrame(tk.Frame):
    """ 
    Manages the view of an optimization by iterating through the positions of the particles.
    """
    def __init__(self, initialize_window: callable, change_menu: callable, optimization: Optimization, function_fig: Figure, bg: str = Color.test3_bg) -> "ViewFrame":
        super().__init__(bg=bg)
        self.__initialize_window: callable = initialize_window
        self.__change_menu: callable = change_menu
        self.__optimization: Optimization = optimization
        self.__function_fig: Figure = function_fig
        back_button_image: tk.PhotoImage = tk.PhotoImage(file="graphics/assets/arrow-back.png").subsample(4)
        back_button_active_image: tk.PhotoImage = tk.PhotoImage(file="graphics/assets/arrow-back-active.png").subsample(4)
        self.__title: tk.Label = tk.Label(self, text=f"Optimization {optimization.index}", height=2, bg=Color.select_title_bg, fg=Color.select_title_fg, font=font.Font(family=FontName.title, size=15))
        self.__back_button: BackButton = BackButton(self, image=back_button_image, active_image=back_button_active_image, width=50, height=50, change_menu=self.back_button_callable, change_menu_args={"menu_name": "select"})
        # self.__pause_button # * Could be a BackButton object
        # self.__reset_button # * Could be a BackButton object
        self.__fig_main_axis: Axes = self.__function_fig.get_axes()[0]
        self.__iteration_canvases: list[tk.Canvas] = [self.__create_iteration_canvas(iteration) for iteration in range(0, optimization.iterations + 1)]
        self.__exited_menu: bool = False

    def back_button_callable(self, **args) -> None:
        self.__exited_menu = True
        self.after(500, self.__change_menu(**args)) # * allows display_iteration_canvases to reach another iteration so the condition of exited_menu is checked

    def __create_iteration_canvas(self, iteration: int) -> tk.Canvas:
        fig: Figure = deepcopy(self.__function_fig)
        fig.set_size_inches(4.8, 4.8)
        fig_main_axis: Axes = fig.get_axes()[0] # * Main axes: title, x-axis and y-axis
        fig_main_axis.set_title(f"{self.__optimization.function_choice} - Iteration {iteration}")
        iteration_df_start: int = (iteration * self.__optimization.swarm.get_particle_amount()) + iteration
        iteration_df_end: int = iteration_df_start + self.__optimization.swarm.get_particle_amount()
        iteration_df: pd.DataFrame = self.__optimization.optimization_df["Position"].iloc[iteration_df_start:iteration_df_end]
        for coordinates in iteration_df:
            fig_main_axis.scatter(coordinates[0], coordinates[1], color="red")
        figure_canvas: FigureCanvasTkAgg = FigureCanvasTkAgg(figure=fig, master=self)
        figure_canvas.draw()
        return figure_canvas.get_tk_widget()
    
    def display_iteration_canvas(self, iteration: int) -> None:
        if self.__exited_menu:
            return None
        try:
            self.__iteration_canvases[iteration-1].pack_forget()
        except IndexError:
            pass
        finally:
            self.__iteration_canvases[iteration].pack(fill="none", expand=True)
            if iteration < self.__optimization.iterations:
                self.after(1000, self.display_iteration_canvas, iteration + 1) # * Schedules the next iteration

    def display(self) -> None:
        for iteration_canvas in self.__iteration_canvases:
            iteration_canvas.pack_forget()
        self.__back_button.focus_set()
        self.__exited_menu = False
        self.__initialize_window(width=500, height=600, title=
        "View optimization - PSO")
        self.__title.pack(fill="x", anchor="e")
        self.__back_button.display(x=0, y=0)
        self.place(x=0, y=0, width=500, height=600)
        self.display_iteration_canvas(0)

    def forget(self) -> None:
        self.__exited_menu = True
        self.place_forget()



import tkinter as tk
from tkinter import font

from psopackage.graphics.colors import Color
from psopackage.graphics.fonts import FontName
from psopackage.graphics.optionsButton import OptionsButton

class CreateButton(OptionsButton):
    """
    CreateButton is a custom button class that inherits from OptionsButton.
    It toggles between two texts and calls corresponding functions when clicked.
    """
    def __init__(self, parent_frame: tk.Frame, text1: str, text2: str, callable1: callable, callable2: callable, padx: int, pady: int) -> None:
        """
        Initializes the CreateButton with the given parameters.
        
        Args:
            parent_frame (tk.Frame): The parent frame in which the button is placed.
            text1 (str): The initial text of the button.
            text2 (str): The text to toggle to when the button is clicked.
            callable1 (callable): The function to call when the button toggles to text2.
            callable2 (callable): The function to call when the button toggles to text1.
            padx (int): The horizontal padding of the button.
            pady (int): The vertical padding of the button.
        """
        # Initialize the parent class with the first text and other parameters
        super().__init__(parent_frame, text=text1, button_font=font.Font(family=FontName.button, size=10), padx=padx, pady=pady, callable=None, callable_args=None) # ? Callable is none here?
        
        # Raise an error if the two texts are the same
        if text1 == text2:
            raise NotImplementedError("text1 and text2 cannot be the same")
        
        # Store the texts and callables
        self.text1: str = text1
        self.text2: str = text2
        self.__active_text: str = text1
        self.__callable1: callable = callable1
        self.__callable2: callable = callable2
    
    def _release(self, event: tk.Event) -> None:
        """
        Handles the button release event. Toggles the button text and calls the corresponding function.
        
        Args:
            event (tk.Event): The event object containing event data.
        """
        # Check the event type
        if event.type == '3':
            pass
        elif event.type == '5':
            # Change the button colors on release
            self.config(bg=Color.optim_button_abg, fg=Color.back_button_afg, activebackground=Color.optim_button_abg, activeforeground=Color.optim_button_afg)
        
        # Toggle the button text and call the corresponding function
        if self.__active_text == self.text1:
            self.config(text=self.text2)
            self.__active_text = self.text2
            self.__callable1()
        elif self.__active_text == self.text2:
            self.config(text=self.text1)
            self.__active_text = self.text1
            self.__callable2()
        else:
            raise NotImplementedError("The text of the button doesn't match neither callable")


import os

import tkinter as tk
from tkinter import font

from psopackage.graphics.backButton import BackButton
from psopackage.graphics.colors import Color
from psopackage.graphics.fonts import FontName
from psopackage.optimization import Optimization
from psopackage.graphics.optionsButton import OptionsButton
from psopackage.graphics.selectMenu.optimizationFrame import OptimizationFrame
from psopackage.graphics.optionsButton import OptionsButton

class SelectMenu():
    """ 
    Manages the Select Optimization menu.
    """
    def __init__(self, parent_frame: tk.Frame, initialize_window: callable, change_menu: callable, optimization_history: list[Optimization], window_width: int, window_height: int):
        self.__optimization_history: list[Optimization] = optimization_history
        self.__optimization_history.reverse()
        self.__window_width: int = window_width
        self.__window_height: int = window_height
        self.__title_height: int = 2 # * Text units, not pixels. It doesn't correspond exactly to the size of the font. Default is 17, Ubuntu size 15 is 15 + 9 = 24 px.
        self.initialize_window: callable = initialize_window
        self.change_menu: callable = change_menu
        self.root: tk.Frame = tk.Frame(parent_frame, bg=Color.test1_bg)

        self.__title: tk.Label = tk.Label(self.root, text="Select a previous optimization", height=self.__title_height, bg=Color.select_title_bg, fg=Color.select_title_fg, font=font.Font(family=FontName.title, size=15))
        self.__canvas: tk.Canvas = tk.Canvas(self.root, bg=Color.test1_bg)
        container_frame_y_padding: int = 30
        self.__container_frame_width: int = self.__window_width - container_frame_y_padding
        self.__container_frame: tk.Frame = tk.Frame(self.__canvas, bg=Color.test3_bg) # * Will be the basis for or the root of the canvas containing optimization frames and the scrollbar
        self.__scrollbar_width: int = 20
        self.__scrollbar: tk.Scrollbar = tk.Scrollbar(self.root, orient="vertical", command=self.__canvas.yview)
        back_button_image: tk.PhotoImage = tk.PhotoImage(file="graphics/assets/arrow-back.png").subsample(4)
        back_button_active_image: tk.PhotoImage = tk.PhotoImage(file="graphics/assets/arrow-back-active.png").subsample(4)
        self.__back_button: BackButton = BackButton(self.root, image=back_button_image, active_image=back_button_active_image, width=self.__title_height*25, height=self.__title_height*25, change_menu=change_menu)
        self.__no_optimizations_label: tk.Label = tk.Label(self.root, height=2, text="No optimizations have been made yet.", bg=Color.select_label_no_optim_bg, fg=Color.select_label_no_optim_fg)
        self.__create_optimization_button: OptionsButton = OptionsButton(self.root, text="Create optimization", callable=change_menu, callable_args={"menu_name": "create"}, padx=(self.__window_width//2 - 250,)*2, pady=(0,0))
        self.__optimization_frames: list[OptimizationFrame] = []

    def __scroll_mouse_wheel(self, event: tk.Event):
        if os.name == "nt":
            self.__canvas.yview_scroll(int(-1*(event.delta/120)), "units")
        elif os.name == "posix":
            if event.num == 4:
                self.__canvas.yview_scroll(-1, "units")
            elif event.num == 5:
                self.__canvas.yview_scroll(1, "units")

    def display(self):
        self.__title_height = 2 # * To avoid recording the value from the previous display call
        container_frame_height: int = 20 + (len(self.__optimization_frames) * 140) # * Bottom 'padding' of the frame
        self.initialize_window(width=self.__window_width, height=self.__window_height, title="Select optimization - PSO")
        self.__title.pack(fill="x", anchor="e")
        self.__back_button.display(x=0, y=0)
        self.__back_button.focus_set()
        # print(self.root.focus_get())

        if len(self.__optimization_history) == 0:
            print("len 0")
            self.__no_optimizations_label.pack(fill="both", anchor="center", pady=(20, 0), padx=(self.__window_width//2) - 250)
            self.__create_optimization_button.pack_display(fill="x", anchor="center")

        else:
            self.__no_optimizations_label.pack_forget()
            self.__create_optimization_button.forget()
            for optim_frame in self.__optimization_frames:
                optim_frame.view_button.view_frame.forget()
            index = len(self.__optimization_frames)
            while index < len(self.__optimization_history):
                optimization = self.__optimization_history[index]
                self.__optimization_frames.append(OptimizationFrame(self.__container_frame, optimization=optimization, forget_select_menu=self.forget, initialize_window=self.initialize_window, change_menu=self.change_menu, width=self.__container_frame_width - 50, scrollbar_width=self.__scrollbar_width, height=120, separation=20, frame_index=index))
                container_frame_height += 140
                index += 1
            for optim_frame in self.__optimization_frames:
                for child in optim_frame.root.winfo_children() + [optim_frame.root]:
                    child.bind("<MouseWheel>", self.__scroll_mouse_wheel)
                    child.bind("<Button-4>", self.__scroll_mouse_wheel)
                    child.bind("<Button-5>", self.__scroll_mouse_wheel)
            self.__container_frame.bind("<MouseWheel>", self.__scroll_mouse_wheel)
            self.__container_frame.bind("<Button-4>", self.__scroll_mouse_wheel)
            self.__container_frame.bind("<Button-5>", self.__scroll_mouse_wheel)
            self.__title_height *= 25 # * To compensate for the difference between pixels and font size
            self.__scrollbar.place(x=self.__window_width - (15 + self.__scrollbar_width), y=self.__title_height + 15, height=self.__window_height - (self.__title_height + 30), width=self.__scrollbar_width)
            self.__canvas.configure(yscrollcommand=self.__scrollbar.set)
            for frame in self.__optimization_frames:
                frame.display(self.__container_frame_width)
            self.__canvas.create_window((0, 0), window=self.__container_frame, anchor="nw", width=self.__container_frame_width, height=container_frame_height) # * Creates a window inside the current tk.Tk() window, so it only displays a certain portion of the widgets contained inside.
            self.__canvas.configure(scrollregion=self.__canvas.bbox("all"))
            self.__canvas.place(y=(self.__title_height) + 15, x=15, width=self.__window_width - (30 + self.__scrollbar_width), height=self.__window_height - (self.__title_height + 30))
            self.__canvas.yview_moveto(0) # * So that every time the menu is displayed it is showing the top optimizations and not the ones the user was seeing before they went back to the main menu.
        self.root.place(x=0, y=0, width=self.__window_width, height=self.__window_height)

    def forget(self):
        self.root.place_forget()
        self.__canvas.place_forget()



import tkinter as tk
from tkinter import font

from psopackage.graphics.optionsButton import OptionsButton
from psopackage.graphics.colors import Color

class OptionsFrame(tk.Frame):
    """
    OptionsFrame is a custom frame class that contains the main menu options.
    """
    def __init__(self, parent_frame: tk.Frame, change_menu: callable,  window_width: int, window_height: int, title_height, bottom_frame_height: int):
        super().__init__(parent_frame, bg=Color.window_bg)
        self.__window_width: int = window_width
        self.__window_height: int = window_height
        self.__title_height: int = title_height
        self.__bottom_frame_height: int = bottom_frame_height
        BUTTON_PADDING: int = 6
        self.__create_button: OptionsButton = OptionsButton(self, text="Create optimization", callable=change_menu, callable_args={"menu_name": "create"}, padx=BUTTON_PADDING*4, pady=(BUTTON_PADDING*2, BUTTON_PADDING))
        self.__select_button: OptionsButton = OptionsButton(self, text="Select optimization", callable=change_menu, callable_args={"menu_name": "select"}, padx=BUTTON_PADDING*4, pady=BUTTON_PADDING)
        self.__exit_button: OptionsButton = OptionsButton(self, text="Exit optimization", callable=change_menu, callable_args={"menu_name": "exit"}, padx=BUTTON_PADDING*4, pady=(BUTTON_PADDING, BUTTON_PADDING*2))
        self.__buttons = [self.__create_button, self.__select_button, self.__exit_button]

    def display(self):
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)
        self.rowconfigure(2, weight=1)
        self.columnconfigure(0, weight=1)
        self.__create_button.focus_set()
        button_row = 0
        for button in self.__buttons:
            button.grid_display(row=button_row, column=0, sticky="nsew")
            button_row += 1
        # ! CHECK: Functionality of the binding in Windows
        # * Checked that it works properly in Windows
        frame_width = self.__window_width
        frame_height = self.__window_height - (self.__title_height + self.__bottom_frame_height)
        self.place(x=0, y=self.__title_height, width=frame_width, height=frame_height)




from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib import colors
import numpy as np
import tkinter as tk
from tkinter import font

from psopackage.graphics.backButton import BackButton
from psopackage.graphics.colors import Color
from psopackage.graphics.fonts import FontName
from psopackage.graphics.createMenu.createInput import CreateInput
from psopackage.graphics.createMenu.functionChoiceMenu import FunctionChoiceMenu
from psopackage.graphics.createMenu.createButton import CreateButton
from psopackage.optimization import Optimization
from psopackage.database.data import Data

class CreateMenu:
    """
    This class manages all the inputs, buttons and the graph display for the create optimization menu.
    """
    def __init__(self, master_frame, initialize_window: callable, change_menu: callable, optimization_history: list[Optimization], data: Data, window_width: int, window_height: int):
        self.__initialize_window: callable = initialize_window
        self.__change_menu: callable = change_menu
        self.__optimization_history: list[Optimization] = optimization_history
        self.__data: Data = data
        self.__width: int = window_width
        self.__height: int = window_height
        self.root: tk.Frame = tk.Frame(master_frame)
        # ! Inheritance could also be applied with title -it's literally the same save the text
        self.__title_height: int = 2
        self.__title: tk.Label = tk.Label(self.root, text="Create optimization", font=font.Font(family=FontName.title, size=15), height=self.__title_height, bg=Color.select_title_bg, fg=Color.select_title_fg)
        back_button_image: tk.PhotoImage = tk.PhotoImage(file="graphics/assets/arrow-back.png").subsample(4)
        back_button_active_image: tk.PhotoImage = tk.PhotoImage(file="graphics/assets/arrow-back-active.png").subsample(4)
        self.__back_button: BackButton = BackButton(self.root, image=back_button_image, active_image=back_button_active_image, width=self.__title_height * 25, height=self.__title_height * 25, change_menu=change_menu, change_menu_args={"menu_name": "main"})

        self.__inputs_frame: tk.Frame = tk.Frame(self.root, bg=Color.test1_bg)
        entry_width: int = 20
        self.__w_coefficient_input: CreateInput = CreateInput(self.__inputs_frame, default_value="0.7", text="Intertia coefficiet",width=entry_width)
        self.__cog_coefficient_input: CreateInput = CreateInput(self.__inputs_frame, default_value=2, text="Cognitive coefficient", width=entry_width)
        self.__soc_coefficient_input: CreateInput = CreateInput(self.__inputs_frame, default_value=2, text="Social coefficient", width=entry_width)
        self.__function_option: FunctionChoiceMenu = FunctionChoiceMenu(self.__inputs_frame, text="Choose a function", options=["Sphere", "Booth", "Goldstein-Price", "Rastrigin"], display_graph=self.display_graph)
        self.__particle_amount_input: CreateInput = CreateInput(self.__inputs_frame, default_value="1", text="Particle amount", width=entry_width)
        self.__iterations_input: CreateInput = CreateInput(self.__inputs_frame, default_value="1", text="Iterations", width=entry_width)
        self.__inputs: list[CreateInput] = [self.__w_coefficient_input, self.__cog_coefficient_input, self.__soc_coefficient_input, self.__function_option, self.__particle_amount_input, self.__iterations_input]

        self.__buttons_frame: tk.Frame = tk.Frame(self.root, bg=Color.test2_bg)
        self.__run_view_button: CreateButton = CreateButton(parent_frame=self.__buttons_frame, text1="Run optimization", text2="View optimization", callable1=self.__run_optimization, callable2=self.__view_optimization, padx=20, pady=20)
        self.__reset_button: CreateButton = CreateButton(parent_frame=self.__buttons_frame, text1="Reset", text2="New optimization", callable1=self.__reset_parameters, callable2=self.__reset_parameters, padx=20, pady=20)
        
        self.__figs: dict = {"Sphere": self.__create_fig("Sphere"), "Booth": self.__create_fig("Booth"), "Goldstein-Price": self.__create_fig("Goldstein-Price"), "Rastrigin": self.__create_fig("Rastrigin")}
        self.__plot_canvases: dict = {"Sphere": self.__create_fig_canvas(self.__figs["Sphere"]), "Booth": self.__create_fig_canvas(self.__figs["Booth"]), "Goldstein-Price": self.__create_fig_canvas(self.__figs["Goldstein-Price"]), "Rastrigin": self.__create_fig_canvas(self.__figs["Rastrigin"])}

    def __run_optimization(self) -> None:
        new_optimization = Optimization(index=len(self.__optimization_history) + 1, function_fig=self.__figs[self.__function_option.get_choice()], function_choice=self.__function_option.get_choice(), data=self.__data, cognitive_coefficient=float(self.__cog_coefficient_input.get_input()), inertia_coefficient=float(self.__w_coefficient_input.get_input()), social_coefficient=float(self.__soc_coefficient_input.get_input()), particle_amount=int(self.__particle_amount_input.get_input()), dimensions=3, iterations=int(self.__iterations_input.get_input()))
        new_optimization.optimize()
        self.__optimization_history.append(new_optimization)

    def __reset_parameters(self) -> None:
        # * Resets all the inputs to their default values
        self.__w_coefficient_input.entry.delete(0, "end")
        self.__cog_coefficient_input.entry.delete(0, "end")
        self.__soc_coefficient_input.entry.delete(0, "end")
        self.__particle_amount_input.entry.delete(0, "end")
        self.__iterations_input.entry.delete(0, "end")
        self.__function_option.choice = "Sphere"
        self.__w_coefficient_input.entry.focus_set()
        self.__run_view_button.config(text=self.__run_view_button.text1)

    def __view_optimization(self) -> None:
        self.__change_menu("select")

    def run_or_view_optimization(self, e, create_optimization: bool) -> None:
        if create_optimization == True:
            self.__run_optimization()
        else:
            self.__view_optimization()

    def __create_contour_levels(self, levels_boundaries: list[float]) -> np.linspace:
        # * Creates the levels for the contour plot based on the boundaries of the test functions
        i = 1
        levels: np.ndarray = np.linspace(start=levels_boundaries[0], stop=levels_boundaries[1], num=10) # * So that levels isn't empty and can be concatenated
        while i < len(levels_boundaries) - 1:
            levels = np.concatenate((levels, np.linspace(start=levels_boundaries[i] + (levels_boundaries[i]/10), stop=levels_boundaries[i + 1], num=10)))
            i += 1
        return levels
    
    def __create_x_y_values(self, bound: int) -> tuple[np.ndarray]:
        x: np.ndarray = np.linspace(start=-bound, num=100, stop=bound)
        y: np.ndarray = np.linspace(start=-bound, num=100, stop=bound)
        return np.meshgrid(x, y)

    def __create_fig(self, option: str) -> Figure:
        # * Generates the figure for the contour plot for each test function
        fig = Figure(figsize=(5,5))
        plot = fig.add_subplot(111)
        levels_boundaries: list[float] = []
        z: np.ndarray = None
        contour = None

        if option == "Sphere":
            x, y = self.__create_x_y_values(bound=2)
            z = x**2 + y**2
            levels_boundaries = [0, 1, 2, 3, 4, 5, 6, 7, 8]
            levels = self.__create_contour_levels(levels_boundaries)
            contour = plot.contourf(x, y, z, levels=levels, cmap="viridis") # * Sets the different colors for different z-values.
            plot.set_title("Sphere function")

        if option == "Booth":
            x, y = self.__create_x_y_values(bound=10)
            z = (x + 2*y - 7)**2 + (2*x + y - 5)**2
            levels_boundaries = [0.1, 1, 10, 100, 1000, 10000] # * Ticks or labels in the colorbar following the test functions graphs
            levels = self.__create_contour_levels(levels_boundaries)
            norm = colors.LogNorm(vmin=levels_boundaries[0], vmax=levels_boundaries[len(levels_boundaries) - 1]) # * Applying a logaritmic normalization to the colorbar
            contour = plot.contourf(x, y, z, levels=levels, cmap="viridis", norm=norm)
            plot.set_title("Booth function")

        if option == "Goldstein-Price":
            x, y = self.__create_x_y_values(bound=2)
            z = (1 + (x+y+1)**2 * (19 - 14 * x + 3 * x**2 - 14 * y + 6*x*y + 3*y**2)) * (30 + (2*x - 3*y)**2 * (18 - 32 * x + 12 * x**2 + 48 * y - 36*x*y + 27 * y**2))
            levels_boundaries = [1, 10, 100, 1000, 10000, 100000, 1000000]
            levels = self.__create_contour_levels(levels_boundaries)
            norm = colors.LogNorm(vmin=levels_boundaries[0], vmax=levels_boundaries[len(levels_boundaries) - 1])
            contour = plot.contourf(x, y, z, levels=levels, cmap="viridis", norm=norm)
            plot.set_title("Goldstein-Price function")

        if option == "Rastrigin":
            x, y = self.__create_x_y_values(bound=5.12)
            z = 10*2 + (x**2 - 10*np.cos(2 * np.pi * x)) + (y**2 - 10*np.cos(2 * np.pi * y))
            levels_boundaries = [0, 10, 20, 30, 40, 50, 60, 70, 80]
            levels = self.__create_contour_levels(levels_boundaries)
            contour = plot.contourf(x, y, z, levels=levels, cmap="viridis")
            plot.set_title("Rastrigin function")

        fig.colorbar(contour, ax=plot, ticks=levels_boundaries)
        plot.set_xlabel("X axis")
        plot.set_ylabel("Y axis")
        fig.subplots_adjust(left=0.225, right=0.90, top=0.825, bottom=0.225) # * Centers the graph
        return fig
    
    def __create_fig_canvas(self, fig: Figure) -> tk.Canvas:
        fig_canvas: FigureCanvasTkAgg = FigureCanvasTkAgg(figure=fig, master=self.root)
        fig_canvas.draw()
        return fig_canvas.get_tk_widget()

    def display_graph(self, option: str):
        for canvas in self.__plot_canvases.values():
            canvas.pack_forget()

        self.__plot_canvases[option].pack(fill="x", expand=True, anchor="ne", side="right", padx=10, pady=(10, 10))

    def display(self):
        self.__initialize_window(width=self.__width, height=self.__height, title="Create Optimization (PSO)")
        self.__title.pack(fill="x", anchor="n", side="top")
        self.__back_button.focus_set()
        for row in range(6):
            self.__inputs_frame.grid_rowconfigure(row, weight=1)
        self.__inputs_frame.columnconfigure(0, weight=1)
        self.__inputs_frame.columnconfigure(1, weight=1)
        i = 0
        for input in self.__inputs:
            input.grid(label_row=i, column=0, sticky="nsew")
            i += 1
        self.__inputs_frame.pack(fill="both", expand=True, anchor="sw", side="left", padx=(10, 0), pady=(10, 10))
        self.__buttons_frame.rowconfigure(0, weight=1)
        self.__buttons_frame.columnconfigure(0, weight=1)
        self.__buttons_frame.columnconfigure(1, weight=1)
        self.__run_view_button.grid_display(row=0, column=0, sticky="nsew")
        self.__reset_button.grid_display(row=0, column=1, sticky="nsew")
        self.__buttons_frame.pack(fill="both", expand=True, anchor="se", side="bottom", pady=(0, 10))
        self.__back_button.display(x=0, y=0)
        self.display_graph("Sphere")
        self.root.place(x=0, y=0, width=self.__width, height=self.__height)

    def forget(self) -> None:
        self.root.focus_set()
        self.root.place_forget()
        self.__title.pack_forget()
        self.__inputs_frame.pack_forget()
        self.__buttons_frame.pack_forget()
        self.__back_button.forget()


"""
A module to represent the velocity vector in a Particle Swarm
Optimization (PSO) algorithm.
The Velocity class inherits from the Vector class and provides 
methods for updating the velocity based on the particle's best
position (pbest) and the global best position (gbest).

## Classes
Velocity: Represents the velocity vector in a PSO algorithm.
Child of Vector.

Vector: Inherited from the base_vector module.

### Methods
- _update(w, c1, c2, pbest, gbest): Updates the velocity
vector based on the given parameters.
- Other getters and setters inherited from the 
base_vector module.
"""

import numpy as np

from psopackage.vector.base_vector import Vector

class Velocity(Vector):
    def __init__(self, dimensions: int = 3) -> None:
        super().__init__(dimensions)
    
if __name__ ==  "__main__":
    v = Velocity(3)
    v.initialize_randomly(1)
    print(v.get_coordinates())
    # ! Correct
    # v._update(w=0.7, c1=2.05, c2=2.05, pbest=np.ndarray(np.array(3), ), gbest=Vector(3))
    print(v.get_coordinates())


import os

import numpy as np
import openpyxl as px
import openpyxl.styles as px_styles
import openpyxl.worksheet.worksheet as px_worksheet
import pandas as pd
from psopackage.graphics.colors import Color

class Data:
    """
    This class is used to store the optimization results in an Excel file. It is used to store the history of the particles and the gbest indexes of each optimization.
    
    """
    def __init__(self, excel_file_name: str) -> None:
        """
        Initializes the Data object with the given Excel file name.
        
        Args:
            excel_file_name (str): The name of the Excel file (without extension and directory).
        """    

        # List to store the history of particles for each optimization
        self.__particle_history: list[pd.DataFrame] = []
        
        # List to store the gbest indexes for each optimization
        self.__gbest_history: list[list[int]] = []
        
        # Counter to keep track of the number of optimizations
        self.__number_of_optimizations: int = 0
        
        # Name of the Excel file (without extension)
        self.__xlsx_name: str = excel_file_name
        
        # Full path to the Excel file
        self.__xlsx_path: str = f"database/optimization_results/{self.__xlsx_name}.xlsx"
        
        # * The two dots are needed if the GUI is directly executed.
        # ! For now, the execution will continue to be done in the gui.py file, but the final version MUST CHANGE the paths to execute everything from the main.py file.
        # * Already done
        
    def append_gbest_indexes(self, optimization_gbest_indexes: list[int]) -> None:
        """
        The gbest indexes are appended to the gbest history list. This method is called after each optimization.
        
        """
        self.__gbest_history.append(optimization_gbest_indexes)
        self.__number_of_optimizations += 1
    
    def append_optimization(self, optimization_df: pd.DataFrame) -> None:
        """
        This method styles the DataFrame and appends it to the Excel file. Creates the Excel file if it does not exist.
        
        """
        # TODO: Test the whole class with multiple sessions and files. Also, update and add documentation and the class diagram.
        # * Seems to be working alright
        if "optimization_results" not in os.listdir("database"):
            os.mkdir("database/optimization_results")
            
        # ? We may want to implement this using with
        
        # * Creating the Excel file if it does not exist
        if not os.path.exists(self.__xlsx_path):
            self.create_spreadsheet()
        
        # * Open the workbook and CREATE THE SHEET
        workbook: px.Workbook = px.load_workbook(self.__xlsx_path)
        sheet_number: int = self.__number_of_optimizations
        sheet_name: str = f"Optimization {sheet_number}"
        if sheet_name not in workbook.sheetnames:
            workbook.create_sheet(title=f"Optimization {sheet_number}")
        else:
            pass

        sheet: px_worksheet.Worksheet = workbook[sheet_name]
        # * Freezing the first two written rows
        # ? Should it only be the second one?
        sheet.freeze_panes = "A3"
        sheet.freeze_panes = "A4"

        # * Variables for styling the BORDERS
        title_side = px_styles.Side(style="slantDashDot", color=Color.siders) 
        # ! Replace for a Color attribute once the branch is merged
        # * Done
        title_border = px_styles.Border(left=title_side, right=title_side,
            top=title_side, bottom=title_side)
        header_side = px_styles.Side(style="thick", color=Color.siders) 
        # ! Replace for a Color attribute once the branch is merged
        # * Done
        header_border = px_styles.Border(left=header_side,
            right=header_side, top=header_side, bottom=header_side)
        common_side = px_styles.Side(style="thin", color=Color.common_side) 
        # ! Replace for a Color attribute once the branch is merged
        # * Done
        common_border = px_styles.Border(left=common_side,
            right=common_side, top=common_side, bottom=common_side)
        spacing_side = px_styles.Side(style="thin", color=Color.siders) 
        # ! Replace for a Color attribute once the branch is merged
        # * Done
        spacing_border = px_styles.Border(left=spacing_side,
            right=spacing_side, top=common_side, bottom=common_side)
        
        # * Setting the same FONT and size for all the cells
        # * It is only overwrittern in the headers (because of the font color)
        for row in sheet.iter_rows(min_row=3, max_row=optimization_df.shape[0] + 3, min_col=2, max_col=7):
            sheet.row_dimensions[row[0].row].height = 18
            for cell in row:
                cell.font = px_styles.Font(name="FreeMono", size=11,
                    bold=True)
                        
        # * Setting and styling the TITLE
        sheet.merge_cells("B2:G2")
        sheet.row_dimensions[2].height = 45
        sheet["B2"] = f"PSO optimization {self.__number_of_optimizations} results"
        title_cell = sheet.cell(row=2, column=2)
        title_cell.alignment = px_styles.Alignment(horizontal="center",
            vertical="center")
        title_cell.font = px_styles.Font(name="FreeMono", size=18,
            bold=True, color=Color.title_cell) 
        # ! Replace for a Color attribute once the branch is merged
        # * Done
        title_cell.fill = px_styles.GradientFill(stop=(Color.grad_fill_start, Color.grad_fill_stop),
            type="linear", degree=90) 
        # ! Replace for a Color attribute once the branch is merged. AND CHECK THE GRADIENT
        # * Done
        title_cell.border = title_border
        
        # * Naming the columns by naming and styling their HEADERS
        sheet["B3"] = "Iteration"
        sheet["C3"] = "Particle"
        sheet["D3"] = "Heuristic coordinates"
        sheet["E3"] = "Position coordinates"
        sheet["F3"] = "Velocity coordinates"
        sheet["G3"] = "Best personal position coordinates"
        for column in range(2, 8):
            cell = sheet.cell(row=3, column=column)
            cell.alignment = px_styles.Alignment(horizontal="center",
                vertical="center", wrap_text=True)
            cell.font = px_styles.Font(name="FreeMono", size=11, bold=True, color= Color.siders)
            cell.fill = px_styles.PatternFill(start_color= Color.cell_fill, end_color= Color.cell_fill, fill_type="solid") 
            # ! @U Replace for a Color attribute once the branch is merged
            # * Done
            cell.border = header_border
            sheet.row_dimensions[3].height = 40

        # * Converting the DataFrame to a tuple of tuples to access the data more easily
        optimization_records: tuple[tuple] = optimization_df.to_records(index=False)

        # * Setting the WIDTH of the columns according to the dimesion of the particles
        column_width: int = 6*(len(optimization_records[0][0]) + 1)
        for column in ["C", "D", "E", "F", "G"]:
            sheet.column_dimensions[column].width = column_width
        # number_of_iterations: int = optimization_df["Heuristic"].isna().sum() # ! Seems it is not needed
        sheet.column_dimensions["B"].width = 10

        # * Couting the AMOUNT OF PARTICLES
        number_of_particles: int = 0
        while optimization_records[number_of_particles][0] is not np.nan:
            number_of_particles += 1
        particle_index: int = 1
        # * Setting the particle indexes in the "C" column and styling them
        for row in range(optimization_df.shape[0]):
            cell = sheet.cell(row=row + 4, column=3)
            if optimization_records[row][0] is not np.nan:
                cell.value = particle_index
                cell.alignment = px_styles.Alignment(horizontal="center",
                    vertical="center")
                cell.fill = px_styles.PatternFill(start_color= Color.cell2_fill, 
                    end_color= Color.cell2_fill, fill_type="solid") 
                # ! Replace for a Color attribute once the branch is merged
                # * Done
                cell.border = common_border
                if particle_index == number_of_particles:
                    particle_index = 1
                else:
                    particle_index += 1
            else:
                # * Styling the spacing rows
                cell.border = spacing_border

        # * Merging the "B" column cells to show the ITERATION number
        starting_row: int = 4
        ending_row: str = number_of_particles + 3
        iteration_number: int = 0
        while starting_row < optimization_df.shape[0]:
            sheet.merge_cells(f"B{starting_row}:B{ending_row}")
            sheet[f"B{starting_row}"] = iteration_number
            sheet.cell(row=starting_row, column=2).fill = px_styles.PatternFill(start_color= Color.cell2_fill, end_color= Color.cell2_fill,
                fill_type="solid") 
            # ! Replace for a Color attribute once the branch is merged
            # * Done
            sheet.cell(row=starting_row, column=2).border = common_border
            sheet.cell(row=ending_row, column=2).border = common_border
            sheet[f"B{starting_row}"].alignment = px_styles.Alignment(horizontal="center", vertical="center")
            sheet.cell(row=ending_row + 1, column=2).border = spacing_border
            starting_row = ending_row + 2
            ending_row = starting_row + number_of_particles - 1
            iteration_number += 1
            # ! Checked the style of the last row
            # * Looks good 

        # * Appending the ARRAYS OF COORDINATES to the corresponding cell and styling them
        for row in range(optimization_df.shape[0]):
            for column in range(optimization_df.shape[1]):
                value = optimization_records[row][column]
                cell = sheet.cell(row=row + 4, column=column + 4)
                if isinstance(value, np.ndarray):
                    # * To avoid iterating through NaNs
                    value: str = ', '.join(map(str, value))
                    # * Makes each value of the tuple of floats a string (using map) and then joins them with a comma in a single string
                    cell.fill = px_styles.PatternFill(start_color=Color.cell2_fill, end_color=Color.cell2_fill, fill_type="solid") 
                    # ! Replace for a Color attribute once the branch is merged
                    # * Done
                    cell.border = common_border
                else:
                    cell.border = spacing_border
                cell.value = value
                cell.alignment = px_styles.Alignment(horizontal="center",
                    vertical="center", wrap_text=True)
                # * Rows and columns start at 1

        workbook.save(self.__xlsx_path)
        self.__particle_history.append(optimization_df)

    def create_spreadsheet(self) -> None:
        try:
            # * Initializing file as an Excel one
            empty_df: pd.DataFrame = pd.DataFrame()
            with pd.ExcelWriter(self.__xlsx_path, engine="openpyxl") as writer:
                empty_df.to_excel(excel_writer=writer)
                writer.book.active.title = "Optimization 1"
                writer.book.save(self.__xlsx_path)
        except FileExistsError as e:
            print(e)
            print("No need to create it again (should happen only in testing).")

    def print_optimization(self, optimization_index: int) -> None:
        pd.set_option("display.max_columns", None)
        pd.set_option("display.expand_frame_repr", False)
        pd.set_option("max_colwidth", None)
        pd.set_option("display.max_rows", None)
        print(self.__particle_history[optimization_index])
        
    def get_particle_history(self) -> list[pd.DataFrame]:
        return self.__particle_history


"""
A module to represent the position vector in a Particle Swarm
Optimization (PSO) algorithm.
The Position class inherits from the Vector class and provides
methods for updating the position based on the velocity vector.

## Classes
Position: Represents the position vector in a PSO algorithm.
Child of Vector.

Vector: Inherited from the base_vector module.

Velocity: Inherited from the velocity module.

### Methods
- _update(velocity): Updates the position based on the given velocity
vector.
- Other getters and setters inherited from the base_vector module.
"""

from psopackage.vector.base_vector import Vector
from psopackage.vector.velocity import Velocity 
import numpy as np

# ! Is a position vector really needed? Could  it be only a Heuristic?
class Position(Vector):
    """
    Represents the position of an object in a multi-dimensional space.

    ## Parameters
    - dimensions : int, optional
        The number of dimensions of the vector. Default is 3.

    ## Attributes
    - _color : dict
        The color of the vector (inherited).
    - _coordinates: np.ndarray
        The coordinates of the vector (inherited).
    - _dimensions : int
        The number of dimensions of the vector (inherited).
    - _coordinates : ndarray
        The coordinates of the position vector.

    ## Methods
    - _update(velocity)
        Updates the position by adding the coordinates of the given 
        velocity vector.
    - Other getters and setters inherited.
    """

    def __init__(self, dimensions: int = 3) -> None:
        super().__init__(dimensions)
    
    def _update(self, velocity: Velocity) -> None:
        # ? Should it be public?
        """
        Updates the position by adding the coordinates of the given
        velocity vector.

        ## Parameters

        - velocity : Velocity
            The velocity vector that contains the coordinates to be added to
            the position.
        """
        self._coordinates += velocity.get_coordinates()

if __name__ ==  "__main__":
    p = Position(3)
    p.initialize_randomly(1)
    print(p.get_coordinates())
    v = Velocity(3)
    v.initialize_randomly(1)
    print(v.get_coordinates())
    p._update(v)
    print(p.get_coordinates())


import os
import time

import numpy as np
import tkinter as tk
from tkinter import font

from psopackage.graphics.colors import Color
from psopackage.graphics.createMenu.createMenu import CreateMenu
from psopackage.graphics.exitMenu import ExitMenu
from psopackage.graphics.mainMenu.mainMenu import MainMenu
from psopackage.graphics.selectMenu.selectMenu import SelectMenu
from psopackage.optimization import Optimization
from psopackage.database.data import Data

class GUI:
    __root: tk.Tk = tk.Tk()
    def __init__(self, optimization_history: list[Optimization], data: Data = Data("test"), program_version: str = "Error") -> None:
        # ? Future versions could include thread management. Could be an interesting way to start learning about parallelism and concurrency.
        self._master_frame: tk.Frame = tk.Frame(GUI.__root, bg=Color.test2_bg)
        self._window_height: int = 0
        self._window_width: int = 0
        self.__create_menu: CreateMenu = CreateMenu(self._master_frame, initialize_window=self._initialize_root, change_menu=self._change_menu, optimization_history=optimization_history, data=data, window_width=750, window_height=500)
        self.__exit_menu: ExitMenu = ExitMenu(self._master_frame, self._initialize_root)
        self.__main_menu: MainMenu = MainMenu(self._master_frame, self._initialize_root, self._change_menu, program_version)
        self.__select_menu: SelectMenu = SelectMenu(self._master_frame, self._initialize_root, self._change_menu, optimization_history, window_width=750, window_height=500)
        self.__menus: dict = {"exit": self.__exit_menu, "main": self.__main_menu, "select": self.__select_menu, "create": self.__create_menu}
        self.optimization_history = optimization_history

    # ! Consider organizing the module files in module packages so graphics/ isn't too messy
    def _change_menu(self, menu_name="") -> None:
        if menu_name in self.__menus:
            for menu in self.__menus.values():
                menu.forget()
            self.__menus[menu_name].display()
            print(menu_name)
            # print(self.optimization_history)
        else:
            raise Exception(f"Menu {menu_name} not found.")

    def _initialize_root(self, width: int, height: int, title: str = "Particle Swarm Optimization (PSO)") -> None:
        GUI.__root.title(title)

        # * Setting initial geometry (dimensions)
        GUI.__root.resizable(False, False)
        screen_width: int = GUI.__root.winfo_screenwidth()
        screen_height: int= GUI.__root.winfo_screenheight()
        self.__window_width: int = width
        self.__window_height: int = height
        top_left_x: int = (screen_width // 2) - (self.__window_width // 2)
        top_left_y: int = (screen_height // 2) - (self.__window_height // 2)
        GUI.__root.geometry(f"{self.__window_width}x{self.__window_height}+{top_left_x}+{top_left_y}")
        self._master_frame.place(x=0, y=0, width=width, height=height)

        # * Setting icon for the application switcher, the dock and the taskbar (Windows)
        small_logo_path: str = "graphics/assets/logo.png"
        large_logo_path: str = small_logo_path
        small_logo: tk.PhotoImage = tk.PhotoImage(file=small_logo_path).subsample(10)
        large_logo: tk.PhotoImage = tk.PhotoImage(file=large_logo_path)
        GUI.__root.iconphoto(False, small_logo, large_logo)

        # * Setting background color
        GUI.__root.configure(bg=Color.window_bg)

    def run(self):
        self._change_menu("main")
        print(GUI.__root.mainloop())

if __name__ == "__main__":
    gui = GUI(optimization_history=[], program_version="0.9.9")
    gui.run()


import tkinter as tk
from tkinter import font

from psopackage.graphics.colors import Color
from psopackage.graphics.fonts import FontName

class OptionsButton(tk.Button):
    """ 
    Manages the buttons in the options frame. 
    """
    def __init__(self, parent_frame: tk.Frame, text: str, callable: callable, callable_args: dict, padx: tuple, pady: tuple, button_font: font.Font = None, args: dict = {
    "bg": Color.optim_button_bg,
    "fg": Color.optim_button_fg,
    "relief": "flat",
    "highlightbackground": Color.optim_button_hbg,
    "highlightcolor": Color.optim_button_hcolor,
    "highlightthickness": 1,
    "cursor": "hand2",
    }) -> "OptionsButton":
        # ! button_font should be font
        # TODO: Make the color-related arguments attribute arguments so that it does depend in the initialization in case the default values are different (and for the mere purpose of keeping the code consistent and good)
        # * Done
        if button_font == None:
            args.update({"font": font.Font(family=FontName.button, size=10)})
        super().__init__(parent_frame, text=text, **args)
        self.__parent_frame: tk.Frame = parent_frame
        self.__callable: callable = callable
        self.__callable_args: dict = callable_args
        self._padx: tuple = padx
        self._pady: tuple = pady
        self.optim_button_fg = Color.optim_button_fg
        self.optim_button_hbg = Color.optim_button_hbg
        self.optim_button_hcolor = Color.optim_button_hcolor
        self.optim_button_abg = Color.optim_button_abg
        self.optim_button_bg = Color.optim_button_bg
        self.back_button_afg = Color.back_button_afg

    def __enter(self, e) -> None:
        self.config(bg=Color.optim_button_abg, fg=Color.back_button_afg, activebackground=Color.optim_button_abg, activeforeground=Color.optim_button_afg)

    def __leave(self, e) -> None:
        self.config(bg=Color.optim_button_bg, fg=Color.optim_button_fg)

    def __click(self, e) -> None:
        self.config(bg=Color.optim_button_cbg, fg=Color.optim_button_cfg, activebackground=Color.optim_button_cbg, activeforeground=Color.optim_button_cfg)
    
    def _release(self, event: tk.Event) -> None:
        if event.type == '3':
            pass
        elif event.type == '5':
            self.config(bg=Color.optim_button_abg, fg=Color.back_button_afg, activebackground=Color.optim_button_abg, activeforeground=Color.optim_button_afg) # ! Consider changing the colors to instance attributes
        self.__parent_frame.forget()
        self.__callable(**self.__callable_args)

    def bind_to_events(self) -> None:
        self.bind("<Enter>", self.__enter)
        self.bind("<Leave>", self.__leave)
        self.bind("<Button-1>", self.__click)
        self.bind("<ButtonRelease-1>", self._release)
        self.bind("<KeyRelease-Return>", self._release)

    def grid_display(self, row: int, column: int, sticky: str) -> None:
        self.bind_to_events()
        self.grid(row=row, column=column, padx=self._padx, pady=self._pady, sticky=sticky)

    def pack_display(self, fill: str, anchor: str) -> None:
        self.bind_to_events()
        self.pack(fill=fill, anchor=anchor, padx=self._padx, pady=self._pady)



"""
This module provides a Heuristic class that represents a vector with a 
heuristic value.

## Classes
Heuristic: A class that represents a vector with a heuristic value.

Vector: Inherited from the base_vector module

### Methods
_update (position): Updates the vector based on the given position vector.

Other methods inherited from the Vector class of the base_vector module.

#### Getters
get_heuristic_f(): Returns the heuristic function

Other getters inherited from the Vector class of the base_vector module.
"""

import numpy as np
import math
from psopackage.vector.base_vector import Vector
from psopackage.vector.position import Position

class Heuristic(Vector):
    """
    Represents a heuristic vector used in Particle Swarm Optimization.

    ## Parameters
    - dimensions : int, optional
        The number of dimensions of the vector. Default is 3.
    - heuristic : callable, optional
        The heuristic function used to calculate the heuristic value.
        Default is `default_heuristic`.

    ## Attributes
    - _color : dict
        The color of the vector (inherited).
    - _coordinates: np.ndarray
        The coordinates of the vector (inherited).
    - _dimensions : int
        The number of dimensions of the vector (inherited).
    - _heuristic : callable
        The heuristic function used to calculate the heuristic value.

    ## Methods
    - _update(self, position)
        Updates the vector based on the given position (overriden).
    - get_heuristic_f(self) -> callable
        Returns the heuristic function used by the vector.
    - Other getters and setters inherited.
    """

    def __init__(self, heuristic_f, dimensions=3):
        super().__init__(dimensions=dimensions)
        self._heuristic_f: callable = heuristic_f
    
    def _update(self, position: Position):
        """
        Updates the vector based on the given position and the 
        heuristic function

        ## Parameters
        position : Position
            The position object used to update the vector.
        """
        heuristic_value: Position = self._heuristic_f(position) # TODO: Static type to be defined
        # ? I think Position works
        new_coordinates: np.ndarray = self.get_coordinates().copy()
        for i in range(self.get_dimensions() - 1):
            new_coordinates[i] = position.get_coordinates().copy()[i]
        new_coordinates[self.get_dimensions() - 1] = heuristic_value
        self.set_coordinates(new_coordinates)
    
    def get_heuristic_f(self) -> callable:
        return self._heuristic_f

if __name__ == "__main__":
    h = Heuristic(3)
    h.initialize_randomly(10)
    print(h.get_coordinates())
    vect = Vector(2)
    vect.initialize_randomly(10)
    print(vect)
    h._update(vect)
    print(h.get_coordinates())



import tkinter as tk

from psopackage.graphics.colors import Color
from psopackage.graphics.selectMenu.viewFrame import ViewFrame
from psopackage.optimization import Optimization

class ViewButton(tk.Button):
    """ 
    Sets the button to change the windonw and show the optimization.
    """
    def __init__(self, master: tk.Frame, image: tk.Image, active_image: tk.Image, forget_select_menu: callable, initialize_window: callable, change_menu: callable,optimization: Optimization, bg: str = Color.select_label_optim_bg, fg: str=Color.select_label_optim_fg) -> "ViewButton":
        super().__init__(master=master, image=image, bg=bg, fg=fg)
        self.__image: tk.Image = image
        self.__active_image: tk.Image = active_image
        self.view_frame: ViewFrame = ViewFrame(optimization=optimization, initialize_window=initialize_window, change_menu=change_menu, function_fig=optimization.function_fig)
        self.__forget_select_menu: callable = forget_select_menu
        self.__bind_to_events()

    def __enter(self, e):
        self.config(bg=Color.preview_button_abg, image=self.__active_image)

    def __leave(self, e):
        self.config(bg=Color.preview_button_bg, image=self.__image)

    def __click(self, e):
        self.config(bg=Color.preview_button_cbg, image=self.__image)

    def __release(self, e):
        self.__forget_select_menu()
        self.view_frame.display()
    
    def __bind_to_events(self) -> None:
        self.bind("<Enter>", self.__enter)
        self.bind("<Leave>", self.__leave)
        self.bind("<Button-1>", self.__click)
        self.bind("<ButtonRelease-1>", self.__release)
        self.bind("<KeyRelease-Return>", self.__release)